# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Tests for global LP/NLP in the MindtPy solver."""

import pyomo.common.unittest as unittest
from pyomo.contrib.mindtpy.tests.eight_process_problem import EightProcessFlowsheet
from pyomo.contrib.mindtpy.tests.nonconvex1 import Nonconvex1
from pyomo.contrib.mindtpy.tests.nonconvex2 import Nonconvex2
from pyomo.contrib.mindtpy.tests.nonconvex3 import Nonconvex3
from pyomo.contrib.mindtpy.tests.nonconvex4 import Nonconvex4
from pyomo.environ import SolverFactory, value
from pyomo.opt import TerminationCondition
from pyomo.contrib.mcpp import pyomo_mcpp

required_solvers = ('baron', 'cplex_persistent')
if not all(SolverFactory(s).available(exception_flag=False) for s in required_solvers):
    subsolvers_available = False
elif not SolverFactory('baron').license_is_valid():
    subsolvers_available = False
else:
    subsolvers_available = True


model_list = [
    EightProcessFlowsheet(convex=False),
    Nonconvex1(),
    Nonconvex2(),
    Nonconvex3(),
    Nonconvex4(),
]


@unittest.skipIf(
    not subsolvers_available,
    'Required subsolvers %s are not available' % (required_solvers,),
)
@unittest.skipIf(not pyomo_mcpp.mcpp_available(), 'MC++ is not available')
class TestMindtPy(unittest.TestCase):
    """Tests for the MindtPy solver plugin."""

    def check_optimal_solution(self, model, places=1):
        for var in model.optimal_solution:
            self.assertAlmostEqual(
                var.value, model.optimal_solution[var], places=places
            )

    def test_GOA(self):
        """Test the global outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='GOA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    single_tree=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=2
                )
                self.check_optimal_solution(model)

    def test_GOA_tabu_list(self):
        """Test the global outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='GOA',
                    mip_solver=required_solvers[1],
                    nlp_solver=required_solvers[0],
                    single_tree=True,
                    use_tabu_list=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=2
                )
                self.check_optimal_solution(model)

    @unittest.skipUnless(
        SolverFactory('gurobi_persistent').available(exception_flag=False)
        and SolverFactory('gurobi_direct').available(),
        'gurobi_persistent and gurobi_direct solver is not available',
    )
    def test_GOA_Gurobi(self):
        """Test the global outer approximation decomposition algorithm."""
        with SolverFactory('mindtpy') as opt:
            for model in model_list:
                model = model.clone()
                results = opt.solve(
                    model,
                    strategy='GOA',
                    mip_solver='gurobi_persistent',
                    nlp_solver=required_solvers[0],
                    single_tree=True,
                )

                self.assertIn(
                    results.solver.termination_condition,
                    [TerminationCondition.optimal, TerminationCondition.feasible],
                )
                self.assertAlmostEqual(
                    value(model.objective.expr), model.optimal_value, places=2
                )
                self.check_optimal_solution(model)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.contrib.appsi.cmodel import cmodel_available
from pyomo.common.gsl import find_GSL


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestIpoptPersistent(unittest.TestCase):
    def test_external_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find the amplgls.dll library')

        opt = pyo.SolverFactory('appsi_ipopt')
        if not opt.available(exception_flag=False):
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.hypot = pyo.ExternalFunction(library=DLL, function='gsl_hypot')
        m.x = pyo.Var(bounds=(-10, 10), initialize=2)
        m.y = pyo.Var(initialize=2)
        e = 2 * m.hypot(m.x, m.x * m.y)
        m.c = pyo.Constraint(expr=e == 2.82843)
        m.obj = pyo.Objective(expr=m.x)
        res = opt.solve(m)
        pyo.assert_optimal_termination(res)
        self.assertAlmostEqual(pyo.value(m.c.body) - pyo.value(m.c.lower), 0)

    def test_external_function_in_objective(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find the amplgls.dll library')

        opt = pyo.SolverFactory('appsi_ipopt')
        if not opt.available(exception_flag=False):
            raise unittest.SkipTest

        m = pyo.ConcreteModel()
        m.hypot = pyo.ExternalFunction(library=DLL, function='gsl_hypot')
        m.x = pyo.Var(bounds=(1, 10), initialize=2)
        m.y = pyo.Var(bounds=(1, 10), initialize=2)
        e = 2 * m.hypot(m.x, m.x * m.y)
        m.obj = pyo.Objective(expr=e)
        res = opt.solve(m)
        pyo.assert_optimal_termination(res)
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 1)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from typing import TYPE_CHECKING, Any, Literal, overload

import pyomo.environ as _environ

if TYPE_CHECKING:
    import pyomo.contrib.appsi.base as _appsi
    import pyomo.contrib.solver.common.factory as _contrib
    import pyomo.opt.base.solvers as _solvers

__doc__ = """
Preview capabilities through ``pyomo.__future__``
=================================================

This module provides a uniform interface for gaining access to future
("preview") capabilities that are either slightly incompatible with the
current official offering, or are still under development with the
intent to replace the current offering.

Currently supported ``__future__`` offerings include:

.. autosummary::

   solver_factory

"""

solver_factory_v1: "_solvers.SolverFactoryClass"
solver_factory_v2: "_appsi.SolverFactoryClass"
solver_factory_v3: "_contrib.SolverFactoryClass"


def __getattr__(name):
    if name in ("solver_factory_v1", "solver_factory_v2", "solver_factory_v3"):
        return solver_factory(int(name[-1]))
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


@overload
def solver_factory(version: None = None) -> int: ...
@overload
def solver_factory(version: Literal[1]) -> "_solvers.SolverFactoryClass": ...
@overload
def solver_factory(version: Literal[2]) -> "_appsi.SolverFactoryClass": ...
@overload
def solver_factory(version: Literal[3]) -> "_contrib.SolverFactoryClass": ...
@overload
def solver_factory(version: int) -> Any: ...


def solver_factory(version: int | None = None) -> int | Any:
    """Get (or set) the active implementation of the SolverFactory

    This allows users to query / set the current implementation of the
    SolverFactory that should be used throughout Pyomo.  Valid options are:

    - ``1``: the original Pyomo SolverFactory
    - ``2``: the SolverFactory from APPSI
    - ``3``: the SolverFactory from pyomo.contrib.solver

    The current active version can be obtained by calling the method
    with no arguments

    .. doctest::

        >>> from pyomo.__future__ import solver_factory
        >>> solver_factory()
        1

    The active factory can be set either by passing the appropriate
    version to this function:

    .. doctest::

        >>> solver_factory(3)
        <pyomo.contrib.solver.common.factory.SolverFactoryClass object ...>

    or by importing the "special" name:

    .. doctest::

        >>> from pyomo.__future__ import solver_factory_v3

    .. doctest::
       :hide:

       >>> from pyomo.__future__ import solver_factory_v1

    """
    global _active_solver_factory_version
    import pyomo.opt.base.solvers as _solvers
    import pyomo.contrib.solver.common.factory as _contrib
    import pyomo.contrib.appsi.base as _appsi

    versions = {
        1: _solvers.LegacySolverFactory,
        2: _appsi.SolverFactory,
        3: _contrib.SolverFactory,
    }
    # The user is just asking what the current SolverFactory is; tell them.
    if version is None:
        if "_active_solver_factory_version" not in globals():
            for ver, cls in versions.items():
                if cls._cls is _environ.SolverFactory._cls:
                    _active_solver_factory_version = ver
                    break
        return _active_solver_factory_version
    #
    # Update the current SolverFactory to be a shim around (shallow copy
    # of) the new active factory
    src = versions.get(version, None)
    if src is not None:
        _active_solver_factory_version = version
        for attr in ('_description', '_cls', '_doc'):
            setattr(_environ.SolverFactory, attr, getattr(src, attr))
    else:
        raise ValueError(
            "Invalid value for target solver factory version; expected {1, 2, 3}, "
            f"received {version}"
        )
    return src


_active_solver_factory_version = solver_factory()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: UTF-8 -*-
"""Transformation to remove zero terms from constraints."""

from pyomo.core import quicksum
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.transformation import TransformationFactory
import pyomo.core.expr as EXPR
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.repn import generate_standard_repn
from pyomo.common.config import ConfigDict, ConfigValue


@TransformationFactory.register(
    'contrib.remove_zero_terms', doc="Remove terms 0 * var in constraints"
)
class RemoveZeroTerms(IsomorphicTransformation):
    """Looks for :math:`0 v` in a constraint and removes it.

    Currently limited to processing linear constraints of the form :math:`x_1 =
    0 x_3`, occurring as a result of fixing :math:`x_2 = 0`.

    .. note:: TODO: support nonlinear expressions

    """

    CONFIG = ConfigDict("RemoveZeroTerms")
    CONFIG.declare(
        "constraints_modified",
        ConfigValue(
            default={},
            description="A dictionary that maps the constraints modified during "
            "the transformation to a tuple: (original_expr, modified_expr)",
        ),
    )

    def _apply_to(self, model, **kwargs):
        """Apply the transformation."""
        config = self.CONFIG(kwargs)
        m = model

        for constr in m.component_data_objects(
            ctype=Constraint, active=True, descend_into=True
        ):
            repn = generate_standard_repn(constr.body)
            if not repn.is_linear() or repn.is_constant():
                continue  # we currently only process linear constraints, and we
                # assume that trivial constraints have already been
                # deactivated or will be deactivated in a different
                # step

            original_expr = constr.expr
            # get the index of all nonzero coefficient variables
            nonzero_vars_idx = [
                i
                for i, _ in enumerate(repn.linear_vars)
                if not repn.linear_coefs[i] == 0
            ]
            const = repn.constant

            # reconstitute the constraint, including only variable terms with
            # nonzero coefficients
            constr_body = (
                quicksum(
                    repn.linear_coefs[i] * repn.linear_vars[i] for i in nonzero_vars_idx
                )
                + const
            )
            if constr.equality:
                new_expr = constr_body == constr.upper
            elif constr.has_lb() and not constr.has_ub():
                new_expr = constr_body >= constr.lower
            elif constr.has_ub() and not constr.has_lb():
                new_expr = constr_body <= constr.upper
            else:
                # constraint is a bounded inequality of form a <= x <= b.
                new_expr = EXPR.inequality(constr.lower, constr_body, constr.upper)
            constr.set_value(new_expr)
            config.constraints_modified[constr] = (original_expr, new_expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import gc
from io import StringIO
import weakref

from pyomo.common.unittest import TestCase
from pyomo.common.log import LoggingIntercept
from pyomo.common.plugin_base import (
    Interface,
    Plugin,
    SingletonPlugin,
    ExtensionPoint,
    implements,
    alias,
    PluginFactory,
    PluginError,
    PluginGlobals,
    DeprecatedInterface,
)


class TestPlugin(TestCase):
    def test_plugin_interface(self):
        class IFoo(Interface):
            pass

        class myFoo(Plugin):
            implements(IFoo)

        ep = ExtensionPoint(IFoo)
        self.assertEqual(ep.extensions(), [])
        self.assertEqual(IFoo._plugins, {myFoo: {}})
        self.assertEqual(len(ep), 0)

        a = myFoo()
        self.assertEqual(ep.extensions(), [])
        self.assertEqual(IFoo._plugins, {myFoo: {0: (weakref.ref(a), False)}})
        self.assertEqual(len(ep), 0)

        a.activate()
        self.assertEqual(ep.extensions(), [a])
        self.assertEqual(IFoo._plugins, {myFoo: {0: (weakref.ref(a), True)}})
        self.assertEqual(len(ep), 1)

        a.deactivate()
        self.assertEqual(ep.extensions(), [])
        self.assertEqual(IFoo._plugins, {myFoo: {0: (weakref.ref(a), False)}})
        self.assertEqual(len(ep), 0)

        # Free a and make sure the garbage collector collects it (so
        # that the weakref will be removed from IFoo._plugins)
        a = None
        gc.collect()
        gc.collect()
        gc.collect()

        self.assertEqual(ep.extensions(), [])
        self.assertEqual(IFoo._plugins, {myFoo: {}})
        self.assertEqual(len(ep), 0)

    def test_singleton_plugin_interface(self):
        class IFoo(Interface):
            pass

        class mySingleton(SingletonPlugin):
            implements(IFoo)

        ep = ExtensionPoint(IFoo)
        self.assertEqual(ep.extensions(), [])
        self.assertEqual(
            IFoo._plugins,
            {mySingleton: {0: (weakref.ref(mySingleton.__singleton__), False)}},
        )
        self.assertIsNotNone(mySingleton.__singleton__)

        with self.assertRaisesRegex(
            RuntimeError, 'Cannot create multiple singleton plugin instances'
        ):
            mySingleton()

        class myDerivedSingleton(mySingleton):
            pass

        self.assertEqual(ep.extensions(), [])
        self.assertEqual(
            IFoo._plugins,
            {
                mySingleton: {0: (weakref.ref(mySingleton.__singleton__), False)},
                myDerivedSingleton: {
                    1: (weakref.ref(myDerivedSingleton.__singleton__), False)
                },
            },
        )
        self.assertIsNotNone(myDerivedSingleton.__singleton__)
        self.assertIsNot(mySingleton.__singleton__, myDerivedSingleton.__singleton__)

        class myDerivedNonSingleton(mySingleton):
            __singleton__ = False

        self.assertEqual(ep.extensions(), [])
        self.assertEqual(
            IFoo._plugins,
            {
                mySingleton: {0: (weakref.ref(mySingleton.__singleton__), False)},
                myDerivedSingleton: {
                    1: (weakref.ref(myDerivedSingleton.__singleton__), False)
                },
                myDerivedNonSingleton: {},
            },
        )
        self.assertIsNone(myDerivedNonSingleton.__singleton__)

        class myServiceSingleton(mySingleton):
            implements(IFoo, service=True)

        self.assertEqual(ep.extensions(), [myServiceSingleton.__singleton__])
        self.assertEqual(
            IFoo._plugins,
            {
                mySingleton: {0: (weakref.ref(mySingleton.__singleton__), False)},
                myDerivedSingleton: {
                    1: (weakref.ref(myDerivedSingleton.__singleton__), False)
                },
                myDerivedNonSingleton: {},
                myServiceSingleton: {
                    2: (weakref.ref(myServiceSingleton.__singleton__), True)
                },
            },
        )
        self.assertIsNotNone(myServiceSingleton.__singleton__)

    def test_inherit_interface(self):
        class IFoo(Interface):
            def fcn(self):
                return 'base'

            def baseFcn(self):
                return 'baseFcn'

        class myFoo(Plugin):
            implements(IFoo)

        with self.assertRaises(AttributeError):
            myFoo().fcn()

        class myFoo(Plugin):
            implements(IFoo, inherit=True)

            def fcn(self):
                return 'derived'

        self.assertEqual(myFoo().fcn(), 'derived')
        self.assertEqual(myFoo().baseFcn(), 'baseFcn')

        class IMock(Interface):
            def mock(self):
                return 'mock'

        class myCombined(myFoo):
            implements(IMock, inherit=True)

        a = myCombined()
        self.assertEqual(a.fcn(), 'derived')
        self.assertEqual(a.baseFcn(), 'baseFcn')
        self.assertEqual(a.mock(), 'mock')

    def test_plugin_factory(self):
        class IFoo(Interface):
            pass

        ep = ExtensionPoint(IFoo)

        class myFoo(Plugin):
            implements(IFoo)
            alias('my_foo', 'myFoo docs')

        factory = PluginFactory(IFoo)
        self.assertEqual(factory.services(), ['my_foo'])

        self.assertIsInstance(factory('my_foo'), myFoo)
        self.assertIsNone(factory('unknown'), None)

        self.assertEqual(factory.doc('my_foo'), 'myFoo docs')
        self.assertEqual(factory.doc('unknown'), '')

        self.assertIs(factory.get_class('my_foo'), myFoo)
        self.assertIsNone(factory.get_class('unknown'))

        a = myFoo()
        b = myFoo()

        self.assertFalse(a.enabled())
        self.assertFalse(b.enabled())
        self.assertIsNone(ep.service())

        a.activate()
        self.assertTrue(a.enabled())
        self.assertFalse(b.enabled())
        self.assertIs(ep.service(), a)

        factory.deactivate('my_foo')
        self.assertFalse(a.enabled())
        self.assertFalse(b.enabled())
        self.assertIsNone(ep.service())

        b.activate()
        self.assertFalse(a.enabled())
        self.assertTrue(b.enabled())
        self.assertIs(ep.service(), b)

        # Note: Run the GC to ensure the instance created by
        # factory('my_foo') above has been removed.
        gc.collect()
        gc.collect()
        gc.collect()

        factory.activate('my_foo')
        self.assertTrue(a.enabled())
        self.assertTrue(b.enabled())
        with self.assertRaisesRegex(
            PluginError,
            r"The ExtensionPoint does not have a unique service!  "
            r"2 services are defined for interface 'IFoo' \(key=None\).",
        ):
            self.assertIsNone(ep.service())

        a.deactivate()
        self.assertFalse(a.enabled())
        self.assertTrue(b.enabled())
        self.assertIs(ep.service(), b)

        factory.activate('unknown')
        self.assertFalse(a.enabled())
        self.assertTrue(b.enabled())
        self.assertIs(ep.service(), b)

        factory.deactivate('unknown')
        self.assertFalse(a.enabled())
        self.assertTrue(b.enabled())
        self.assertIs(ep.service(), b)

    def test_deprecation(self):
        out = StringIO()
        with LoggingIntercept(out):
            PluginGlobals.add_env(None)
        self.assertIn(
            "Pyomo only supports a single global environment",
            out.getvalue().replace('\n', ' '),
        )

        out = StringIO()
        with LoggingIntercept(out):
            PluginGlobals.pop_env()
        self.assertIn(
            "Pyomo only supports a single global environment",
            out.getvalue().replace('\n', ' '),
        )

        out = StringIO()
        with LoggingIntercept(out):
            PluginGlobals.clear()
        self.assertIn(
            "Pyomo only supports a single global environment",
            out.getvalue().replace('\n', ' '),
        )

        class IFoo(Interface):
            pass

        out = StringIO()
        with LoggingIntercept(out):

            class myFoo(Plugin):
                alias('myFoo', subclass=True)

        self.assertIn(
            "alias() function does not support the subclass",
            out.getvalue().replace('\n', ' '),
        )

        out = StringIO()
        with LoggingIntercept(out):

            class myFoo(Plugin):
                implements(IFoo, namespace='here')

        self.assertIn(
            "only supports a single global namespace.",
            out.getvalue().replace('\n', ' '),
        )

        class IGone(DeprecatedInterface):
            __deprecated_version__ = '1.2.3'
            pass

        out = StringIO()
        with LoggingIntercept(out):

            class myFoo(Plugin):
                implements(IGone)

        self.assertIn(
            "The IGone interface has been deprecated", out.getvalue().replace('\n', ' ')
        )

        out = StringIO()
        with LoggingIntercept(out):
            ExtensionPoint(IGone).extensions()
        self.assertIn(
            "The IGone interface has been deprecated", out.getvalue().replace('\n', ' ')
        )

        class ICustomGone(DeprecatedInterface):
            __deprecated_message__ = 'This interface is gone!'
            __deprecated_version__ = '1.2.3'

        out = StringIO()
        with LoggingIntercept(out):

            class myFoo(Plugin):
                implements(ICustomGone)

        self.assertIn("This interface is gone!", out.getvalue().replace('\n', ' '))

        out = StringIO()
        with LoggingIntercept(out):
            ExtensionPoint(ICustomGone).extensions()
        self.assertIn("This interface is gone!", out.getvalue().replace('\n', ' '))



# -*- coding: utf-8 -*-

# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from pyomo.contrib.mindtpy.config_options import _get_MindtPy_FP_config
from pyomo.contrib.mindtpy.algorithm_base_class import _MindtPyAlgorithm
from pyomo.core import ConstraintList
from pyomo.contrib.mindtpy.util import calc_jacobians
from pyomo.opt import SolverFactory
from pyomo.contrib.mindtpy.cut_generation import add_oa_cuts


@SolverFactory.register(
    'mindtpy.fp', doc='MindtPy: Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo'
)
class MindtPy_FP_Solver(_MindtPyAlgorithm):
    """
    Decomposition solver for Mixed-Integer Nonlinear Programming (MINLP) problems.

    The MindtPy (Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo) solver
    applies a variety of decomposition-based approaches to solve Mixed-Integer
    Nonlinear Programming (MINLP) problems.
    This class includes:

    - Feasibility pump (FP)
    """

    CONFIG = _get_MindtPy_FP_config()

    def check_config(self):
        # feasibility pump alone will lead to iteration_limit = 0, important!
        self.config.iteration_limit = 0
        self.config.move_objective = True
        super().check_config()

    def initialize_mip_problem(self):
        """Deactivate the nonlinear constraints to create the MIP problem."""
        super().initialize_mip_problem()
        self.jacobians = calc_jacobians(
            self.mip.MindtPy_utils.nonlinear_constraint_list,
            self.config.differentiate_mode,
        )  # preload jacobians
        self.mip.MindtPy_utils.cuts.oa_cuts = ConstraintList(
            doc='Outer approximation cuts'
        )

    def add_cuts(
        self,
        dual_values,
        linearize_active=True,
        linearize_violated=True,
        cb_opt=None,
        nlp=None,
    ):
        add_oa_cuts(
            self.mip,
            dual_values,
            self.jacobians,
            self.objective_sense,
            self.mip_constraint_polynomial_degree,
            self.mip_iter,
            self.config,
            self.timing,
            cb_opt,
            linearize_active,
            linearize_violated,
        )

    def MindtPy_iteration_loop(self):
        pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Pyomo.DoE was produced under the Department of Energy Carbon Capture Simulation
# Initiative (CCSI), and is copyright (c) 2022 by the software owners:
# TRIAD National Security, LLC., Lawrence Livermore National Security, LLC.,
# Lawrence Berkeley National Laboratory, Pacific Northwest National Laboratory,
# Battelle Memorial Institute, University of Notre Dame,
# The University of Pittsburgh, The University of Texas at Austin,
# University of Toledo, West Virginia University, et al. All rights reserved.
#
# NOTICE. This Software was developed under funding from the
# U.S. Department of Energy and the U.S. Government consequently retains
# certain rights. As such, the U.S. Government has been granted for itself
# and others acting on its behalf a paid-up, nonexclusive, irrevocable,
# worldwide license in the Software to reproduce, distribute copies to the
# public, prepare derivative works, and perform publicly and display
# publicly, and to permit other to do so.
# ____________________________________________________________________________________

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)

from enum import Enum
import itertools
import logging

if scipy_available and numpy_available:
    from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxModel

import pyomo.environ as pyo


class FIMExternalGreyBox(
    ExternalGreyBoxModel if (scipy_available and numpy_available) else object
):
    def __init__(self, doe_object, objective_option="determinant", logger_level=None):
        """
        Grey box model for metrics on the FIM. This methodology reduces
        numerical complexity for the computation of FIM metrics related
        to eigenvalue decomposition.

        Parameters
        ----------
        doe_object:
           Design of Experiments object that contains a built model
           (with sensitivity matrix, Q, and fisher information matrix, FIM).
           The external grey box model will utilize elements of the
           `doe_object` model to build the FIM metric with consistent naming.
        obj_option:
           String representation of the objective option. Current available
           options are: ``determinant`` (D-optimality), ``trace`` (A-optimality),
           ``minimum_eigenvalue`` (E-optimality), ``condition_number``
           (modified E-optimality).
           default: ``determinant``
        logger_level:
           logging level to be specified if different from doe_object's logging level.
           default: None, or equivalently, use the logging level of doe_object.

           NOTE: Use logging.DEBUG for all messages.
        """

        if doe_object is None:
            raise ValueError(
                "DoE Object must be provided to build external grey box of the FIM."
            )

        self.doe_object = doe_object

        # Grab parameter list from the doe_object model
        self._param_names = [i for i in self.doe_object.model.parameter_names]
        self._n_params = len(self._param_names)

        # Check if the doe_object has model components that are required
        # TODO: is this check necessary?
        from pyomo.contrib.doe import ObjectiveLib

        objective_option = ObjectiveLib(objective_option)
        self.objective_option = objective_option

        # Create logger for FIM egb object
        self.logger = logging.getLogger(__name__)

        # If logger level is None, use doe_object's logger level
        if logger_level is None:
            logger_level = doe_object.logger.level

        self.logger.setLevel(level=logger_level)

        # Set initial values for inputs
        # Need a mask structure
        self._masking_matrix = np.triu(np.ones_like(self.doe_object.fim_initial))
        self._input_values = np.asarray(
            self.doe_object.fim_initial[self._masking_matrix > 0], dtype=np.float64
        )
        self._n_inputs = len(self._input_values)

    def _get_FIM(self):
        # Grabs the current FIM subject
        # to the input values.
        # This function currently assumes
        # that we use a lower triangular
        # FIM.
        upt_FIM = self._input_values

        # Create FIM in the correct way
        current_FIM = np.zeros_like(self.doe_object.fim_initial)
        # Utilize upper triangular portion of FIM
        current_FIM[np.triu_indices_from(current_FIM)] = upt_FIM
        # Construct lower triangular using the
        # current upper triangle minus the diagonal.
        current_FIM += current_FIM.transpose() - np.diag(np.diag(current_FIM))

        return current_FIM

    def _reorder_pairs(self, i, j, k, l):
        # Reorders the pairs (i, j) and
        # (k, l) for considering only
        # the symmetric portion of the FIM
        # while calculating the Hessian

        # If the pairs ((i, j), (k, l)) are not
        # in increasing order, we reorder
        # the pairs.
        if i > j:
            if k > l:
                return [j, i, l, k]
            else:
                return [j, i, k, l]
        else:
            if k > l:
                return [i, j, l, k]
        return [i, j, k, l]

    def input_names(self):
        # Cartesian product gives us matrix indices flattened in row-first format
        # Can use itertools.combinations(self._param_names, 2) with added
        # diagonal elements, or do double for loops if we switch to upper triangular
        input_names_list = list(
            itertools.combinations_with_replacement(self._param_names, 2)
        )
        return input_names_list

    def equality_constraint_names(self):
        # TODO: Are there any objectives that will have constraints?
        return []

    def output_names(self):
        # TODO: add output name for the variable. This may have to be
        # an input from the user. Or it could depend on the usage of
        # the ObjectiveLib Enum object, which should have an associated
        # name for the objective function at all times.
        from pyomo.contrib.doe import ObjectiveLib

        if self.objective_option == ObjectiveLib.trace:
            obj_name = "A-opt"
        elif self.objective_option == ObjectiveLib.determinant:
            obj_name = "log-D-opt"
        elif self.objective_option == ObjectiveLib.minimum_eigenvalue:
            obj_name = "E-opt"
        elif self.objective_option == ObjectiveLib.condition_number:
            obj_name = "ME-opt"
        else:
            ObjectiveLib(self.objective_option)
        return [obj_name]

    def set_input_values(self, input_values):
        # Set initial values to be flattened initial FIM (aligns with input names)
        np.copyto(self._input_values, input_values)

    def evaluate_equality_constraints(self):
        # TODO: are there any objectives that will have constraints?
        return None

    def evaluate_outputs(self):
        # Evaluates the objective value for the specified
        # ObjectiveLib type.
        current_FIM = self._get_FIM()

        M = np.asarray(current_FIM, dtype=np.float64).reshape(
            self._n_params, self._n_params
        )

        # Change objective value based on ObjectiveLib type.
        from pyomo.contrib.doe import ObjectiveLib

        if self.objective_option == ObjectiveLib.trace:
            obj_value = np.trace(np.linalg.pinv(M))
        elif self.objective_option == ObjectiveLib.determinant:
            sign, logdet = np.linalg.slogdet(M)
            obj_value = logdet
        elif self.objective_option == ObjectiveLib.minimum_eigenvalue:
            eig, _ = np.linalg.eig(M)
            obj_value = np.min(eig)
        elif self.objective_option == ObjectiveLib.condition_number:
            eig, _ = np.linalg.eig(M)
            obj_value = np.log(np.abs(np.max(eig) / np.min(eig)))
        else:
            ObjectiveLib(self.objective_option)

        return np.asarray([obj_value], dtype=np.float64)

    def finalize_block_construction(self, pyomo_block):
        # Set bounds on the inputs/outputs
        # Set initial values of the inputs/outputs
        # This will depend on the objective used

        # Initialize grey box FIM values
        for ind, val in enumerate(self.input_names()):
            pyomo_block.inputs[val] = self.doe_object.fim_initial[
                self._masking_matrix > 0
            ][ind]

        # Initialize log_determinant value
        from pyomo.contrib.doe import ObjectiveLib

        # Calculate initial values for the output
        output_value = self.evaluate_outputs()[0]

        # Set the value of the output for the given
        # objective function.
        if self.objective_option == ObjectiveLib.trace:
            pyomo_block.outputs["A-opt"] = output_value
        elif self.objective_option == ObjectiveLib.determinant:
            pyomo_block.outputs["log-D-opt"] = output_value
        elif self.objective_option == ObjectiveLib.minimum_eigenvalue:
            pyomo_block.outputs["E-opt"] = output_value
        elif self.objective_option == ObjectiveLib.condition_number:
            pyomo_block.outputs["ME-opt"] = output_value

    def evaluate_jacobian_equality_constraints(self):
        # TODO: Do any objectives require constraints?

        # Returns coo_matrix of the correct shape
        return None

    def evaluate_jacobian_outputs(self):
        # Compute the jacobian of the objective function with
        # respect to the fisher information matrix. Then, return
        # a coo_matrix that aligns with what IPOPT will expect.
        current_FIM = self._get_FIM()

        M = np.asarray(current_FIM, dtype=np.float64).reshape(
            self._n_params, self._n_params
        )

        # TODO: Add inertia correction for
        #       negative/small eigenvalues
        eig_vals, eig_vecs = np.linalg.eig(M)
        if min(eig_vals) <= 1e-3:
            pass

        from pyomo.contrib.doe import ObjectiveLib

        if self.objective_option == ObjectiveLib.trace:
            Minv = np.linalg.pinv(M)
            # Derivative formula of A-optimality
            # is -inv(FIM) @ inv(FIM). Add reference to
            # pyomo.DoE 2.0 manuscript S.I.
            jac_M = -Minv @ Minv
        elif self.objective_option == ObjectiveLib.determinant:
            Minv = np.linalg.pinv(M)
            # Derivative formula derived using tensor
            # calculus. Add reference to pyomo.DoE 2.0
            # manuscript S.I.
            jac_M = 0.5 * (Minv + Minv.transpose())
        elif self.objective_option == ObjectiveLib.minimum_eigenvalue:
            # Obtain minimum eigenvalue location
            min_eig_loc = np.argmin(eig_vals)

            # Grab eigenvector associated with
            # the minimum eigenvalue and make
            # it a matrix. This is so we can
            # use matrix operations later in
            # the code.
            min_eig_vec = np.array([eig_vecs[:, min_eig_loc]])

            # Calculate the derivative matrix.
            # This is the expansion product of
            # the eigenvector we grabbed in
            # the previous line of code.
            jac_M = min_eig_vec * np.transpose(min_eig_vec)
        elif self.objective_option == ObjectiveLib.condition_number:
            # Obtain minimum (and maximum) eigenvalue location(s)
            min_eig_loc = np.argmin(eig_vals)
            max_eig_loc = np.argmax(eig_vals)

            min_eig = np.min(eig_vals)
            max_eig = np.max(eig_vals)

            # Grab eigenvector associated with
            # the min (and max) eigenvalue and make
            # it a matrix. This is so we can
            # use matrix operations later in
            # the code.
            min_eig_vec = np.array([eig_vecs[:, min_eig_loc]])
            max_eig_vec = np.array([eig_vecs[:, max_eig_loc]])

            # Calculate the derivative matrix.
            # Similar to minimum eigenvalue,
            # this computation involves two
            # expansion products.
            min_eig_term = min_eig_vec * np.transpose(min_eig_vec)
            max_eig_term = max_eig_vec * np.transpose(max_eig_vec)

            # Combining the expression
            jac_M = 1 / max_eig * max_eig_term - 1 / min_eig * min_eig_term
        else:
            ObjectiveLib(self.objective_option)

        # We are only using a symmetric, triangular
        # representation of the FIM, so we need
        # to add the off-diagonal elements twice.
        jac_M = 2 * jac_M - np.diag(np.diag(jac_M))
        # Filter the Jacobian, jac_M, using the
        # masking matrix to only select the
        # symmetric, triangular components
        jac_M = jac_M[self._masking_matrix > 0]
        M_rows = np.zeros((len(jac_M.flatten()), 1)).flatten()
        M_cols = np.arange(len(jac_M.flatten()))

        # Returns coo_matrix of the correct shape
        return scipy.sparse.coo_matrix(
            (jac_M.flatten(), (M_rows, M_cols)), shape=(1, len(jac_M.flatten()))
        )

    # Beyond here is for Hessian information
    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        # TODO: Do any objectives require constraints?
        # Assert lengths match
        self._eq_con_mult_values = np.asarray(
            eq_con_multiplier_values, dtype=np.float64
        )

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        # TODO: Do any objectives require constraints?
        # Assert length matches
        self._output_con_mult_values = np.asarray(
            output_con_multiplier_values, dtype=np.float64
        )

    def evaluate_hessian_equality_constraints(self):
        # Returns coo_matrix of the correct shape
        # No constraints so this returns `None`
        return None

    def evaluate_hessian_outputs(self):
        # Compute the hessian of the objective function with
        # respect to the fisher information matrix. Then, return
        # a coo_matrix that aligns with what IPOPT will expect.
        current_FIM = self._get_FIM()

        M = np.asarray(current_FIM, dtype=np.float64).reshape(
            self._n_params, self._n_params
        )

        # We will store the Hessian values in
        # vectorized (flattened) format. The length
        # of the vectorized Hessian for the symmetric
        # FIM representation scales by the number of
        # unknown parameters.
        hess_array_length = round(
            (((self._n_params + 1) * self._n_params / 2) + 1)
            * (((self._n_params + 1) * self._n_params / 2))
            / 2
        )

        # Initializing lists of the correct length
        # for the hessian values and the row and column
        # of these data in the coo matrix to be returned
        hess_vals = [0] * hess_array_length
        hess_rows = [0] * hess_array_length
        hess_cols = [0] * hess_array_length

        # We are utilizing the symmetric Hessian, but we
        # must consider the contribution from all elements.
        # Therefore, we are required to use the full product
        # space of the parameter names (full FIM) to compute
        # the Hessian of the symmetric FIM.
        full_input_names = itertools.product(self._param_names, repeat=2)

        # Here, we use combination with replacement to only
        # consider the upper triangle of the Hessian for the
        # full FIM. We will map these second derivative values
        # back onto the symmetric FIM Hessian.
        input_differentials_2D = itertools.combinations_with_replacement(
            full_input_names, 2
        )

        from pyomo.contrib.doe import ObjectiveLib

        if self.objective_option == ObjectiveLib.trace:
            # Grab Inverse
            Minv = np.linalg.pinv(M)

            # Also grab inverse squared
            Minv_sq = Minv @ Minv

            for current_differential in input_differentials_2D:
                d1, d2 = current_differential

                # Grabbing the ordered quadruple (i, j, k, l)
                # `location` here refers to the index in the
                # self._param_names list
                #
                # i is the location of the first element of d1
                # j is the location of the second element of d1
                # k is the location of the first element of d2
                # l is the location of the second element of d2
                i = self._param_names.index(d1[0])
                j = self._param_names.index(d1[1])
                k = self._param_names.index(d2[0])
                l = self._param_names.index(d2[1])

                # New Formula (tested with finite differencing)
                # Will be cited from the Pyomo.DoE 2.0 paper
                hess_contribution = (Minv[i, l] * Minv_sq[k, j]) + (
                    Minv_sq[i, l] * Minv[k, j]
                )

                # Since we are considering the full matrix in
                # this loop, we need to point the contribution
                # to the correct index for the symmetric FIM
                # Hessian.
                reordered_ijkl = self._reorder_pairs(i, j, k, l)
                d1_symmetric = (
                    self._param_names[reordered_ijkl[0]],
                    self._param_names[reordered_ijkl[1]],
                )
                d2_symmetric = (
                    self._param_names[reordered_ijkl[2]],
                    self._param_names[reordered_ijkl[3]],
                )

                # Identify what index of the symmetric FIM
                # Hessian arrays need to be updated.
                # Note: we are only interested in building
                # the lower triangular portion of the Hessian.
                row = max(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                col = min(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                flattened_row_col_index = (row + 1) * row // 2 + col

                # Hessian needs to be handled carefully because of
                # the ``missing`` components from the full FIM
                # when only passing a symmetric version of the FIM.
                #
                # When we reordered (i, j, k, l), we are correctly
                # pointing to which index needs to be contributed to.
                # However, when an element that is not included
                # is being mapped to a diagonal element of the
                # symmetric FIM hessian from the full FIM hessian,
                # it needs to be counted twice. This only occurs
                # when (i != j) and (k != l) and (i, j) and (k, l)
                # are the conjugate of one another:
                # (i == l) and (j == k).
                #
                # Otherwise, we only add the element once.

                # Standard addition
                hess_vals[flattened_row_col_index] += hess_contribution

                # Duplicate check and addition if
                # criteria is satisfied.
                if ((i != j) and (k != l)) and ((i == l) and (j == k)):
                    hess_vals[flattened_row_col_index] += hess_contribution

                hess_rows[flattened_row_col_index] = row
                hess_cols[flattened_row_col_index] = col

        elif self.objective_option == ObjectiveLib.determinant:
            # Grab inverse
            Minv = np.linalg.pinv(M)

            for current_differential in input_differentials_2D:
                # Row, Col and i, j, k, l values are
                # obtained identically as in the trace
                # for loop above.
                d1, d2 = current_differential

                i = self._param_names.index(d1[0])
                j = self._param_names.index(d1[1])
                k = self._param_names.index(d2[0])
                l = self._param_names.index(d2[1])

                # New Formula (tested with finite differencing)
                # Will be cited from the Pyomo.DoE 2.0 paper
                hess_contribution = -(Minv[i, l] * Minv[k, j])

                # Since we are considering the full matrix in
                # this loop, we need to point the contribution
                # to the correct index for the symmetric FIM
                # Hessian.
                reordered_ijkl = self._reorder_pairs(i, j, k, l)
                d1_symmetric = (
                    self._param_names[reordered_ijkl[0]],
                    self._param_names[reordered_ijkl[1]],
                )
                d2_symmetric = (
                    self._param_names[reordered_ijkl[2]],
                    self._param_names[reordered_ijkl[3]],
                )

                # Identify what index of the symmetric FIM
                # Hessian arrays need to be updated
                row = max(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                col = min(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                flattened_row_col_index = (row + 1) * row // 2 + col

                # Hessian needs to be handled carefully because of
                # the ``missing`` components when only passing
                # a symmetric version of the FIM. For a more
                # detailed explanation, please see the trace
                # for loop above
                hess_vals[flattened_row_col_index] += hess_contribution

                # Duplicate check and addition
                if ((i != j) and (k != l)) and ((i == l) and (j == k)):
                    hess_vals[flattened_row_col_index] += hess_contribution

                hess_rows[flattened_row_col_index] = row
                hess_cols[flattened_row_col_index] = col

        elif self.objective_option == ObjectiveLib.minimum_eigenvalue:
            # Grab eigenvalues and eigenvectors
            # Also need the min location
            all_eig_vals, all_eig_vecs = np.linalg.eig(M)
            min_eig_loc = np.argmin(all_eig_vals)

            # Grabbing min eigenvalue and corresponding
            # eigenvector
            min_eig = all_eig_vals[min_eig_loc]
            min_eig_vec = np.array([all_eig_vecs[:, min_eig_loc]])

            for current_differential in input_differentials_2D:
                # Row, Col and i, j, k, l values are
                # obtained identically as in the trace
                # for loop above.
                d1, d2 = current_differential

                i = self._param_names.index(d1[0])
                j = self._param_names.index(d1[1])
                k = self._param_names.index(d2[0])
                l = self._param_names.index(d2[1])

                # For loop to iterate over all
                # eigenvalues/vectors
                hess_contribution = 0
                for curr_eig in range(len(all_eig_vals)):
                    # Skip if we are at the minimum
                    # eigenvalue. Denominator is
                    # zero.
                    if curr_eig == min_eig_loc:
                        continue

                    # Formula derived in Pyomo.DoE Paper
                    hess_contribution += (
                        1
                        * (
                            min_eig_vec[0, i]
                            * all_eig_vecs[j, curr_eig]
                            * min_eig_vec[0, l]
                            * all_eig_vecs[k, curr_eig]
                        )
                        / (min_eig - all_eig_vals[curr_eig])
                    )
                    hess_contribution += (
                        1
                        * (
                            min_eig_vec[0, k]
                            * all_eig_vecs[i, curr_eig]
                            * min_eig_vec[0, j]
                            * all_eig_vecs[l, curr_eig]
                        )
                        / (min_eig - all_eig_vals[curr_eig])
                    )

                # Since we are considering the full matrix in
                # this loop, we need to point the contribution
                # to the correct index for the symmetric FIM
                # Hessian.
                reordered_ijkl = self._reorder_pairs(i, j, k, l)
                d1_symmetric = (
                    self._param_names[reordered_ijkl[0]],
                    self._param_names[reordered_ijkl[1]],
                )
                d2_symmetric = (
                    self._param_names[reordered_ijkl[2]],
                    self._param_names[reordered_ijkl[3]],
                )

                # Identify what index of the symmetric FIM
                # Hessian arrays need to be updated
                row = max(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                col = min(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                flattened_row_col_index = (row + 1) * row // 2 + col

                # Hessian needs to be handled carefully because of
                # the ``missing`` components when only passing
                # a symmetric version of the FIM. See trace for loop
                # for more detailed explanation
                hess_vals[flattened_row_col_index] += hess_contribution

                # Duplicate check and addition
                if ((i != j) and (k != l)) and ((i == l) and (j == k)):
                    hess_vals[flattened_row_col_index] += hess_contribution

                hess_rows[flattened_row_col_index] = row
                hess_cols[flattened_row_col_index] = col

        elif self.objective_option == ObjectiveLib.condition_number:
            # Hessian for log condition number has 4
            # terms. The first and third terms are
            # multiples of the second derivative of the
            # maximum and minimum eigenvalues, respectively
            # The other two are tensor products
            # of the first derivative of the maximum
            # eigenvalue with itself, and the minimum
            # eigenvalue with itself.
            #
            # Grab eigenvalues and eigenvectors
            # Also need the max and min locations
            all_eig_vals, all_eig_vecs = np.linalg.eig(M)
            min_eig_loc = np.argmin(all_eig_vals)
            max_eig_loc = np.argmax(all_eig_vals)

            # Grabbing min eigenvalue and corresponding
            # eigenvector
            min_eig = all_eig_vals[min_eig_loc]
            min_eig_vec = np.array([all_eig_vecs[:, min_eig_loc]])

            # Grabbing max eigenvalue and corresponding
            # eigenvector
            max_eig = all_eig_vals[max_eig_loc]
            max_eig_vec = np.array([all_eig_vecs[:, max_eig_loc]])

            for current_differential in input_differentials_2D:
                # Row, Col and i, j, k, l values are
                # obtained identically as in the trace
                # for loop above.
                d1, d2 = current_differential

                i = self._param_names.index(d1[0])
                j = self._param_names.index(d1[1])
                k = self._param_names.index(d2[0])
                l = self._param_names.index(d2[1])

                # For loop to iterate over all
                # eigenvalues/vectors for first
                # term (second derivative of
                # maximum eigenvalue)
                log_cond_term_1 = 0
                for curr_eig in range(len(all_eig_vals)):
                    # Skip if we are at the maximum
                    # eigenvalue. Denominator is
                    # zero.
                    if curr_eig == max_eig_loc:
                        continue

                    # Formula derived in Pyomo.DoE Paper
                    log_cond_term_1 += (
                        1
                        * (
                            max_eig_vec[0, i]
                            * all_eig_vecs[j, curr_eig]
                            * max_eig_vec[0, l]
                            * all_eig_vecs[k, curr_eig]
                        )
                        / (max_eig - all_eig_vals[curr_eig])
                    )
                    log_cond_term_1 += (
                        1
                        * (
                            max_eig_vec[0, k]
                            * all_eig_vecs[i, curr_eig]
                            * max_eig_vec[0, j]
                            * all_eig_vecs[l, curr_eig]
                        )
                        / (max_eig - all_eig_vals[curr_eig])
                    )

                # For loop to iterate over all
                # eigenvalues/vectors for third
                # term (second derivative of
                # minimum eigenvalue)
                log_cond_term_3 = 0
                for curr_eig in range(len(all_eig_vals)):
                    # Skip if we are at the minimum
                    # eigenvalue. Denominator is
                    # zero.
                    if curr_eig == min_eig_loc:
                        continue

                    # Formula derived in Pyomo.DoE Paper
                    log_cond_term_3 += (
                        1
                        * (
                            min_eig_vec[0, i]
                            * all_eig_vecs[j, curr_eig]
                            * min_eig_vec[0, l]
                            * all_eig_vecs[k, curr_eig]
                        )
                        / (min_eig - all_eig_vals[curr_eig])
                    )
                    log_cond_term_3 += (
                        1
                        * (
                            min_eig_vec[0, k]
                            * all_eig_vecs[i, curr_eig]
                            * min_eig_vec[0, j]
                            * all_eig_vecs[l, curr_eig]
                        )
                        / (min_eig - all_eig_vals[curr_eig])
                    )

                # Computing each term of the hessian formula
                # Second derivative of max eigenvalue term
                log_cond_term_1 = 1 / max_eig * log_cond_term_1

                # First derivative of max eigenvalue term
                log_cond_term_2 = (
                    1
                    / (max_eig**2)
                    * (max_eig_vec[0, l] * max_eig_vec[0, k])
                    * (max_eig_vec[0, j] * max_eig_vec[0, i])
                )

                # Second derivative of min eigenvalue term
                log_cond_term_3 = 1 / min_eig * log_cond_term_3

                # First derivative of min eigenvalue term
                log_cond_term_4 = (
                    1
                    / (min_eig**2)
                    * (min_eig_vec[0, l] * min_eig_vec[0, k])
                    * (min_eig_vec[0, j] * min_eig_vec[0, i])
                )

                # Combining all the components
                hess_contribution = (
                    log_cond_term_1
                    - log_cond_term_2
                    - log_cond_term_3
                    + log_cond_term_4
                )

                # Since we are considering the full matrix in
                # this loop, we need to point the contribution
                # to the correct index for the symmetric FIM
                # Hessian.
                reordered_ijkl = self._reorder_pairs(i, j, k, l)
                d1_symmetric = (
                    self._param_names[reordered_ijkl[0]],
                    self._param_names[reordered_ijkl[1]],
                )
                d2_symmetric = (
                    self._param_names[reordered_ijkl[2]],
                    self._param_names[reordered_ijkl[3]],
                )

                # Identify what index of the symmetric FIM
                # Hessian arrays need to be updated
                row = max(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                col = min(
                    self.input_names().index(d1_symmetric),
                    self.input_names().index(d2_symmetric),
                )
                flattened_row_col_index = (row + 1) * row // 2 + col

                # Hessian needs to be handled carefully because of
                # the ``missing`` components when only passing
                # a symmetric version of the FIM. See trace for loop
                # for more detailed explanation
                hess_vals[flattened_row_col_index] += hess_contribution

                # Duplicate check and addition
                if ((i != j) and (k != l)) and ((i == l) and (j == k)):
                    hess_vals[flattened_row_col_index] += hess_contribution

                hess_rows[flattened_row_col_index] = row
                hess_cols[flattened_row_col_index] = col
        else:
            ObjectiveLib(self.objective_option)

        # Returns coo_matrix of the correct shape
        return scipy.sparse.coo_matrix(
            (np.asarray(hess_vals), (hess_rows, hess_cols)),
            shape=(self._n_inputs, self._n_inputs),
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests explicit bound to variable bound transformation module."""

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Constraint,
    TransformationFactory,
    Var,
    value,
    Integers,
)


class TestConstraintToVarBoundTransform(unittest.TestCase):
    """Tests explicit bound to variable bound transformation."""

    def test_constraint_to_var_bound(self):
        """Test converting explicit constraints into variable bounds."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v3 = Var(initialize=3)
        m.v4 = Var(initialize=4)
        m.v5 = Var(initialize=5)
        m.v6 = Var()
        m.v7 = Var(domain=Integers)
        m.v8 = Var(domain=Integers)
        m.c1 = Constraint(expr=m.v1 == 2)
        m.c2 = Constraint(expr=m.v2 >= -2)
        m.c3 = Constraint(expr=m.v3 <= 5)
        m.c4 = Constraint(expr=m.v4 <= m.v5)
        m.v5.fix()
        m.c6 = Constraint(expr=m.v6 >= 2)
        m.c7 = Constraint(expr=m.v7 + 2 >= 2.01)
        m.c8 = Constraint(expr=m.v8 + 2 >= 2.0001)

        m2 = TransformationFactory('contrib.constraints_to_var_bounds').create_using(m)
        self.assertEqual(value(m2.v1.lb), 2)
        self.assertEqual(value(m2.v1.ub), 2)
        self.assertTrue(m2.v1.fixed)

        self.assertEqual(value(m2.v2.lb), -2)
        self.assertFalse(m2.v2.has_ub())

        self.assertEqual(value(m2.v3.ub), 5)
        self.assertFalse(m2.v3.has_lb())

        self.assertEqual(value(m2.v4.ub), 5)
        self.assertFalse(m2.v4.has_lb())

        self.assertEqual(value(m2.v6.lb), 2)
        self.assertFalse(m2.v6.has_ub())
        self.assertEqual(value(m2.v6, exception=False), None)

        del m2  # to keep from accidentally using it below

        TransformationFactory('contrib.constraints_to_var_bounds').apply_to(
            m, tolerance=1e-3
        )
        self.assertEqual(value(m.v1.lb), 2)
        self.assertEqual(value(m.v1.ub), 2)
        self.assertTrue(m.v1.fixed)

        self.assertEqual(value(m.v2.lb), -2)
        self.assertFalse(m.v2.has_ub())

        self.assertEqual(value(m.v3.ub), 5)
        self.assertFalse(m.v3.has_lb())

        self.assertEqual(value(m.v4.ub), 5)
        self.assertFalse(m.v4.has_lb())

        self.assertEqual(value(m.v7.lb), 1)
        self.assertEqual(value(m.v8.lb), 0)

    def test_skip_trivial_constraints(self):
        """Tests handling of zero coefficients."""
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.z = Var()
        m.c = Constraint(expr=m.x * m.y == m.z)
        m.z.fix(0)
        m.y.fix(0)
        TransformationFactory('contrib.constraints_to_var_bounds').apply_to(m)
        self.assertTrue(m.c.active)
        self.assertFalse(m.x.has_lb())
        self.assertFalse(m.x.has_ub())

    def test_detect_fixed_false(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=m.x == 3)
        TransformationFactory('contrib.constraints_to_var_bounds').apply_to(
            m, detect_fixed=False
        )
        self.assertFalse(m.c.active)
        self.assertTrue(m.x.has_lb())
        self.assertEqual(m.x.lb, 3)
        self.assertTrue(m.x.has_ub())
        self.assertEqual(m.x.ub, 3)
        self.assertFalse(m.x.fixed)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.graphics import matplotlib_available, seaborn_available
from pyomo.contrib.pynumero.asl import AmplInterface
from pyomo.opt import SolverFactory

ipopt_available = SolverFactory("ipopt").available()
pynumero_ASL_available = AmplInterface.available()


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
class TestRooneyBieglerExamples(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    @unittest.skipUnless(pynumero_ASL_available, "test requires libpynumero_ASL")
    @unittest.skipUnless(seaborn_available, "test requires seaborn")
    def test_parameter_estimation_example(self):
        from pyomo.contrib.parmest.examples.rooney_biegler import (
            parameter_estimation_example,
        )

        parameter_estimation_example.main()

    @unittest.skipUnless(seaborn_available, "test requires seaborn")
    def test_bootstrap_example(self):
        from pyomo.contrib.parmest.examples.rooney_biegler import bootstrap_example

        bootstrap_example.main()

    @unittest.skipUnless(seaborn_available, "test requires seaborn")
    def test_likelihood_ratio_example(self):
        from pyomo.contrib.parmest.examples.rooney_biegler import (
            likelihood_ratio_example,
        )

        likelihood_ratio_example.main()


@unittest.skipUnless(pynumero_ASL_available, "test requires libpynumero_ASL")
@unittest.skipUnless(ipopt_available, "The 'ipopt' solver is not available")
@unittest.skipUnless(
    parmest.parmest_available, "Cannot test parmest: required dependencies are missing"
)
class TestReactionKineticsExamples(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    def test_example(self):
        from pyomo.contrib.parmest.examples.reaction_kinetics import (
            simple_reaction_parmest_example,
        )

        simple_reaction_parmest_example.main()


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
class TestSemibatchExamples(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    def test_model(self):
        from pyomo.contrib.parmest.examples.semibatch import semibatch

        semibatch.main()

    def test_parameter_estimation_example(self):
        from pyomo.contrib.parmest.examples.semibatch import (
            parameter_estimation_example,
        )

        parameter_estimation_example.main()

    def test_scenario_example(self):
        from pyomo.contrib.parmest.examples.semibatch import scenario_example

        scenario_example.main()


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
class TestReactorDesignExamples(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    @unittest.pytest.mark.expensive
    def test_model(self):
        from pyomo.contrib.parmest.examples.reactor_design import reactor_design

        reactor_design.main()

    @unittest.skipUnless(pynumero_ASL_available, "test requires libpynumero_ASL")
    def test_parameter_estimation_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import (
            parameter_estimation_example,
        )

        parameter_estimation_example.main()

    @unittest.skipUnless(seaborn_available, "test requires seaborn")
    def test_bootstrap_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import bootstrap_example

        bootstrap_example.main()

    @unittest.pytest.mark.expensive
    def test_likelihood_ratio_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import (
            likelihood_ratio_example,
        )

        likelihood_ratio_example.main()

    @unittest.pytest.mark.expensive
    def test_leaveNout_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import leaveNout_example

        leaveNout_example.main()

    def test_timeseries_data_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import (
            timeseries_data_example,
        )

        timeseries_data_example.main()

    def test_multisensor_data_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import (
            multisensor_data_example,
        )

        multisensor_data_example.main()

    @unittest.skipUnless(
        matplotlib_available and seaborn_available,
        "test requires matplotlib and seaborn",
    )
    def test_datarec_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import datarec_example

        datarec_example.main()

    def test_update_suffix_example(self):
        from pyomo.contrib.parmest.examples.reactor_design import update_suffix_example

        suffix_obj, new_vals, new_var_vals = update_suffix_example.main()

        # Check that the suffix object has been updated correctly
        for i, v in enumerate(new_var_vals):
            self.assertAlmostEqual(new_var_vals[i], new_vals[i], places=6)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Logger function for community_graph.py"""

from logging import getLogger
from pyomo.core import Constraint, Objective, Var
from pyomo.common.dependencies import networkx as nx

logger = getLogger('pyomo.contrib.community_detection')


def _event_log(model, model_graph, constraint_set, type_of_graph, with_objective):
    """
    Logs information about the results of the code in generate_model_graph

    This function takes in the same Pyomo model as generate_model_graph and the model_graph generated by
    generate_model_graph (which is a NetworkX graph of nodes and edges based on the Pyomo model). Then, some relevant
    information about the model and model_graph is determined and logged using the logger.

    This function is designed to be called by generate_model_graph.

    Parameters
    ----------
    model: Block
         the Pyomo model or block to be used for community detection
    model_graph: NetworkX graph
        a NetworkX graph with nodes and edges based on the given Pyomo optimization model
    constraint_set: set
        a set of the numbers that correspond to all of the constraints in the given model
    type_of_graph: str
        a string that specifies the type of graph that is created from the model
        'c' creates a graph based on constraint nodes,
        'v' creates a graph based on variable nodes,
        'b' creates a graph based on constraint and variable nodes (bipartite graph).
    with_objective: bool
        a Boolean argument that specifies whether or not the objective function is included in the graph

    Returns
    -------
    This function returns nothing; it simply logs information that is relevant to the user.
    """

    # Collect some information that will be useful for the logger
    all_variables_count = len(
        list(model.component_data_objects(ctype=Var, descend_into=True))
    )

    active_constraints_count = len(
        list(
            model.component_data_objects(
                ctype=Constraint, active=True, descend_into=True
            )
        )
    )
    all_constraints_count = len(
        list(model.component_data_objects(ctype=Constraint, descend_into=True))
    )

    active_objectives_count = len(
        list(
            model.component_data_objects(
                ctype=Objective, active=True, descend_into=True
            )
        )
    )
    all_objectives_count = len(
        list(model.component_data_objects(ctype=Objective, descend_into=True))
    )

    number_of_nodes, number_of_edges = (
        model_graph.number_of_nodes(),
        model_graph.number_of_edges(),
    )

    # Log this information as info
    logger.info("%s variables found in the model" % all_variables_count)

    logger.info("%s constraints found in the model" % all_constraints_count)
    logger.info("%s active constraints found in the model" % active_constraints_count)

    logger.info("%s objective(s) found in the model" % all_objectives_count)
    logger.info("%s active objective(s) found in the model" % active_objectives_count)

    logger.info("%s nodes found in the graph created from the model" % number_of_nodes)
    logger.info("%s edges found in the graph created from the model" % number_of_edges)

    # Log information on connectivity and density (if density is 1 then that means the nodes are connected
    # to every other node)
    if number_of_nodes > 0:
        if nx.is_connected(model_graph):
            logger.info("The graph created from the model is connected.")
            graph_is_connected = True
        else:
            logger.info("The graph created from the model is disconnected.")
            graph_is_connected = False

        if type_of_graph == 'bipartite':
            if graph_is_connected:
                top_nodes, bottom_nodes = nx.bipartite.sets(model_graph)

                # The purpose of the next few lines is to ensure that (1) top_nodes is not an empty set (as this causes
                # errors) and that (2) the node sets are properly identified as either constraints or variables
                if len(top_nodes) == 0:
                    top_nodes, bottom_nodes = bottom_nodes, top_nodes
                if top_nodes == constraint_set:
                    constraint_nodes, variable_nodes = top_nodes, bottom_nodes
                else:
                    constraint_nodes, variable_nodes = bottom_nodes, top_nodes
            else:
                # If the graph is not connected then we must construct the constraint node set
                # and variable node set manually
                constraint_nodes = {
                    node for node in model_graph.nodes() if node in constraint_set
                }
                variable_nodes = set(model_graph) - constraint_nodes

            constraint_density = round(
                nx.bipartite.density(model_graph, constraint_nodes), 2
            )
            variable_density = round(
                nx.bipartite.density(model_graph, variable_nodes), 2
            )

            if (
                constraint_density == variable_density == 1.0
            ):  # If the graph is complete, both will equal 1
                logger.info(
                    "The bipartite graph constructed from the model is complete (graph density equals 1)"
                )
            else:
                logger.info(
                    "For the bipartite graph constructed from the model, the density for constraint nodes is %s"
                    % constraint_density
                )
                logger.info(
                    "For the bipartite graph constructed from the model, the density for variable nodes is %s"
                    % variable_density
                )

        else:
            graph_density = round(nx.density(model_graph), 2)

            if graph_density == 1.0:
                logger.info(
                    "The graph constructed from the model is complete (graph density equals 1)"
                )
            else:
                logger.info(
                    "The graph constructed from the model has a density of %s"
                    % graph_density
                )

    # Given one of the conditionals below, we will log this information as a warning
    if all_variables_count == 0:
        logger.warning("No variables found in the model")

    if all_constraints_count == 0:
        logger.warning("No constraints found in the model")
    elif active_constraints_count == 0:
        logger.warning("No active constraints found in the model")

    if all_objectives_count == 0:
        if with_objective:
            logger.info(
                "Parameter 'with_objective' is True but no objective(s) found in the model"
            )
        else:
            logger.info("No objective(s) found in the model")
    elif active_objectives_count == 0:
        if with_objective:
            logger.info(
                "Parameter 'with_objective' is True but no active objective(s) found in the model"
            )
        else:
            logger.info("No active objective(s) found in the model")

    if number_of_nodes == 0:
        logger.warning(
            "No nodes were created for the graph (based on the model and the given parameters)"
        )
    if number_of_edges == 0:
        logger.warning(
            "No edges were created for the graph (based on the model and the given parameters)"
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.symbol_map import SymbolMap

from pyomo.core.base.label import TextLabeler


def symbol_map_from_instance(instance):
    """
    Create a symbol map from an instance using name-based labelers.
    """
    from pyomo.core.base import Var, Constraint, Objective

    symbol_map = SymbolMap()
    labeler = TextLabeler()
    #
    # Recursively iterate over all variables
    #
    for varvalue in instance.component_data_objects(Var, active=True):
        symbol_map.getSymbol(varvalue, labeler)
    #
    # Recursively iterate over all constraints
    #
    for constraint_data in instance.component_data_objects(Constraint, active=True):
        con_symbol = symbol_map.getSymbol(constraint_data, labeler)
        if constraint_data.equality:
            label = 'c_e_%s_' % con_symbol
            symbol_map.alias(constraint_data, label)
        else:
            if constraint_data.lower is not None:
                if constraint_data.upper is not None:
                    symbol_map.alias(constraint_data, 'r_l_%s_' % con_symbol)
                    symbol_map.alias(constraint_data, 'r_u_%s_' % con_symbol)
                else:
                    label = 'c_l_%s_' % con_symbol
                    symbol_map.alias(constraint_data, label)
            elif constraint_data.upper is not None:
                label = 'c_u_%s_' % con_symbol
                symbol_map.alias(constraint_data, label)
    #
    # Recursively iterate over all objectives
    #
    first = True
    for objective_data in instance.component_data_objects(Objective, active=True):
        symbol_map.getSymbol(objective_data, labeler)
        if first:
            # The first objective is the default
            symbol_map.alias(objective_data, "__default_objective__")
            first = False
    #
    # Return the symbol map
    #
    return symbol_map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.tempfiles import TempfileManager
import pyomo.environ as pyo
from pyomo.contrib import appsi
from pyomo.contrib.appsi.cmodel import cmodel_available
import os


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestNLWriter(unittest.TestCase):
    def test_all_vars_fixed(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= pyo.exp(m.x))
        m.c2 = pyo.Constraint(expr=m.y >= (m.x - 1) ** 2)
        m.x.fix(1)
        m.y.fix(2)
        writer = appsi.writers.NLWriter()
        with TempfileManager:
            fname = TempfileManager.create_tempfile(suffix='.appsi.nl')
            with self.assertRaisesRegex(
                ValueError, 'there are not any unfixed variables in the problem'
            ):
                writer.write(m, fname)

    def _write_and_check_header(self, m, correct_lines):
        writer = appsi.writers.NLWriter()
        with TempfileManager:
            fname = TempfileManager.create_tempfile(suffix='.appsi.nl')
            writer.write(m, fname)
            with open(fname, 'r') as f:
                for ndx, line in enumerate(list(f.readlines())[:10]):
                    self.assertTrue(line.startswith(correct_lines[ndx]))

    def test_header_1(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c = pyo.Constraint(expr=m.x + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '0 0',
            '0 0',
            '0 0 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_2(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y)
        m.c = pyo.Constraint(expr=m.x + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '0 1',
            '0 0',
            '0 1 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_3(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c = pyo.Constraint(expr=m.x**2 + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 0',
            '0 0',
            '1 0 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_4(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y)
        m.c = pyo.Constraint(expr=m.x**2 + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 1 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_5(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c = pyo.Constraint(expr=m.x**2 + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 2 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_6(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y)
        m.c = pyo.Constraint(expr=m.x**2 + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '2 1 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_7(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c = pyo.Constraint(expr=m.x + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 0',
            '0 0',
            '1 0 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_8(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y)
        m.c = pyo.Constraint(expr=m.x**2 + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 0',
            '0 0',
            '2 0 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_9(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y**2)
        m.c = pyo.Constraint(expr=m.x + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '0 1',
            '0 0',
            '0 1 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_10(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y**2)
        m.c = pyo.Constraint(expr=m.x + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 1 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_11(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y**2)
        m.c = pyo.Constraint(expr=m.x**2 + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 2 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_12(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x + m.y**2)
        m.c = pyo.Constraint(expr=m.x**2 + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '2 1 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_13(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y)
        m.c = pyo.Constraint(expr=m.x + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 2 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_14(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c = pyo.Constraint(expr=m.x + m.y == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '0 1',
            '0 0',
            '0 2 0',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_15(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c = pyo.Constraint(expr=m.x + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '1 2 1',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)

    def test_header_16(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c = pyo.Constraint(expr=m.x**2 + m.y**2 == 1)
        correct_lines = [
            'g3 1 1 0',
            '2 1 1 0 1',
            '1 1',
            '0 0',
            '2 2 2',
            '0 0 0 1',
            '0 0 0 0 0',
            '2 2',
            '0 0',
            '0 0 0 0 0',
        ]
        self._write_and_check_header(m, correct_lines)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

"""A module of utilities for collecting timing information

.. autosummary::

    report_timing
    TicTocTimer
    tic
    toc
    HierarchicalTimer

"""

import functools
import logging
import sys
import time
import traceback
from pyomo.common.deprecation import deprecation_warning
from pyomo.common.modeling import NOTSET as _NotSpecified

_logger = logging.getLogger('pyomo.common.timing')
_logger.propagate = False
_logger.setLevel(logging.WARNING)

_construction_logger = logging.getLogger('pyomo.common.timing.construction')
_transform_logger = logging.getLogger('pyomo.common.timing.transformation')


class report_timing:
    def __init__(self, stream=True, level=logging.INFO):
        """Set reporting of Pyomo timing information.

        For historical reasons, this class may be used as a function
        (the reporting logger is configured as part of the instance
        initializer).  However, the preferred usage is as a context
        manager (thereby ensuring that the timing logger is restored
        upon exit).

        Parameters
        ----------
        stream: bool, TextIOBase

            The destination stream to emit timing information.  If
            ``True``, defaults to ``sys.stdout``.  If ``False`` or
            ``None``, disables reporting of timing information.

        level: int

            The logging level for the timing logger

        """
        self._old_level = _logger.level
        # For historical reasons (because report_timing() used to be a
        # function), we will do what you think should be done in
        # __enter__ here in __init__.
        if stream:
            _logger.setLevel(level)
            if stream is True:
                stream = sys.stdout
            self._handler = logging.StreamHandler(stream)
            self._handler.setFormatter(logging.Formatter("      %(message)s"))
            _logger.addHandler(self._handler)
        else:
            self._handler = list(_logger.handlers)
            _logger.setLevel(logging.WARNING)
            for h in list(_logger.handlers):
                _logger.removeHandler(h)

    def reset(self):
        _logger.setLevel(self._old_level)
        if type(self._handler) is list:
            for h in self._handler:
                _logger.addHandler(h)
        else:
            _logger.removeHandler(self._handler)

    def __enter__(self):
        return self

    def __exit__(self, et, ev, tb):
        self.reset()


class GeneralTimer:
    def __init__(self, fmt, data):
        self.fmt = fmt
        self.data = data

    def report(self):
        _logger.info(self)

    @property
    def obj(self):
        return self.data[-1]

    @property
    def timer(self):
        return self.data[:-1]

    def __str__(self):
        return self.fmt % self.data


class ConstructionTimer:
    __slots__ = ('obj', 'timer')
    msg = "%6.*f seconds to construct %s %s%s"
    in_progress = "ConstructionTimer object for %s %s; %0.3f elapsed seconds"

    def __init__(self, obj):
        self.obj = obj
        self.timer = -default_timer()

    def report(self):
        # Record the elapsed time, as some log handlers may not
        # immediately generate the message string
        self.timer += default_timer()
        _construction_logger.info(self)

    @property
    def name(self):
        try:
            return self.obj.name
        except RuntimeError:
            try:
                return self.obj.local_name
            except RuntimeError:
                return '(unknown)'
        except AttributeError:
            return '(unknown)'

    def __str__(self):
        try:
            if self.obj.is_indexed():
                # indexed component
                if self.obj.index_set().isfinite():
                    idx = len(self.obj.index_set())
                else:
                    idx = len(self.obj)
                idx_label = f'{idx} indices' if idx != 1 else '1 index'
            elif hasattr(self.obj, 'index_set'):
                # scalar indexed components
                idx = len(self.obj.index_set())
                idx_label = f'{idx} indices' if idx != 1 else '1 index'
            else:
                # other non-indexed component (e.g., Suffix)
                idx_label = ''
        except AttributeError:
            # unknown component
            idx_label = ''
        if idx_label:
            idx_label = f'; {idx_label} total'
        try:
            _type = self.obj.ctype.__name__
        except AttributeError:
            _type = type(self.obj).__name__
        total_time = self.timer
        if total_time < 0:
            total_time += default_timer()
            return self.in_progress % (_type, self.name, total_time)
        return self.msg % (
            2 if total_time >= 0.005 else 0,
            total_time,
            _type,
            self.name,
            idx_label,
        )


class TransformationTimer:
    __slots__ = ('obj', 'mode', 'timer')
    msg = "%6.*f seconds to apply Transformation %s%s"
    in_progress = "TransformationTimer object for %s%s; %0.3f elapsed seconds"

    def __init__(self, obj, mode=None):
        self.obj = obj
        if mode is None:
            self.mode = ''
        else:
            self.mode = " (%s)" % (mode,)
        self.timer = -default_timer()

    def report(self):
        # Record the elapsed time, as some log handlers may not
        # immediately generate the message string
        self.timer += default_timer()
        _transform_logger.info(self)

    @property
    def name(self):
        return self.obj.__class__.__name__

    def __str__(self):
        total_time = self.timer
        if total_time < 0:
            total_time += default_timer()
            return self.in_progress % (self.name, self.mode, total_time)
        return self.msg % (
            2 if total_time >= 0.005 else 0,
            total_time,
            self.name,
            self.mode,
        )


#
# Setup the timer
#
# perf_counter is guaranteed to be monotonic and the most accurate
# timer.  It became available in Python 3.3.  Prior to that, clock() was
# more accurate than time() on Windows (.35us vs 15ms), but time() was
# more accurate than clock() on Linux (1ns vs 1us).  It is unfortunate
# that time() is not monotonic, but since the TicTocTimer is used for
# (potentially very accurate) timers, we will sacrifice monotonicity on
# Linux for resolution.
default_timer = time.perf_counter


class TicTocTimer:
    """A class to calculate and report elapsed time.

    Examples:
       >>> from pyomo.common.timing import TicTocTimer
       >>> timer = TicTocTimer()
       >>> timer.tic('starting timer') # starts the elapsed time timer (from 0)
       [    0.00] starting timer
       >>> # ... do task 1
       >>> dT = timer.toc('task 1')
       [+   0.00] task 1
       >>> print("elapsed time: %0.1f" % dT)
       elapsed time: 0.0

    If no ostream or logger is provided, then output is printed to sys.stdout

    Args:
        ostream (FILE): an optional output stream to print the timing
            information
        logger (Logger): an optional output stream using the python
           logging package. Note: the timing logged using ``logger.info()``
    """

    def __init__(self, ostream=_NotSpecified, logger=None):
        if ostream is _NotSpecified and logger is not None:
            ostream = None
        self._lastTime = self._loadTime = default_timer()
        self.ostream = ostream
        self.logger = logger
        self.level = logging.INFO
        self._start_count = 0
        self._cumul = 0

    def tic(
        self,
        msg=_NotSpecified,
        *args,
        ostream=_NotSpecified,
        logger=_NotSpecified,
        level=_NotSpecified,
    ):
        """Reset the tic/toc delta timer.

        This resets the reference time from which the next delta time is
        calculated to the current time.

        Args:
            msg (str): The message to print out.  If not specified, then
                prints out "Resetting the tic/toc delta timer"; if msg
                is None, then no message is printed.
            *args (tuple): optional positional arguments used for
                %-formatting the `msg`
            ostream (FILE): an optional output stream (overrides the ostream
                provided when the class was constructed).
            logger (Logger): an optional output stream using the python
                logging package (overrides the ostream provided when the
                class was constructed). Note: timing logged using logger.info
            level (int): an optional logging output level.

        """
        self._lastTime = self._loadTime = default_timer()
        if msg is _NotSpecified:
            msg = "Resetting the tic/toc delta timer"
        if msg is not None:
            if args and '%' not in msg:
                # Note: specify the parent module scope for the logger
                # so this does not hit (and get handled by) the local
                # pyomo.common.timing logger.
                deprecation_warning(
                    "tic(): 'ostream' and 'logger' should be "
                    "specified as keyword arguments",
                    version='6.4.2',
                    logger=__package__,
                )
                ostream, *args = args
                if args:
                    logger, *args = args
            self.toc(
                msg, *args, delta=False, ostream=ostream, logger=logger, level=level
            )

    def toc(
        self,
        msg=_NotSpecified,
        *args,
        delta=True,
        ostream=_NotSpecified,
        logger=_NotSpecified,
        level=_NotSpecified,
    ):
        """Print out the elapsed time.

        This resets the reference time from which the next delta time is
        calculated to the current time.

        Args:
            msg (str): The message to print out.  If not specified, then
                print out the file name, line number, and function that
                called this method; if `msg` is None, then no message is
                printed.
            *args (tuple): optional positional arguments used for
                %-formatting the `msg`
            delta (bool): print out the elapsed wall clock time since
                the last call to :meth:`tic` (``False``) or since the
                most recent call to either :meth:`tic` or :meth:`toc`
                (``True`` (default)).
            ostream (FILE): an optional output stream (overrides the ostream
                provided when the class was constructed).
            logger (Logger): an optional output stream using the python
                logging package (overrides the ostream provided when the
                class was constructed). Note: timing logged using `level`
            level (int): an optional logging output level.
        """

        # Note: important to do this first so that we don't add a random
        # amount of time for I/O operations or extracting the stack.
        # This helps ensure that the timing tests are less fragile.
        now = default_timer()

        if msg is _NotSpecified:
            msg = 'File "%s", line %s in %s' % traceback.extract_stack(limit=2)[0][:3]
        if args and msg is not None and '%' not in msg:
            # Note: specify the parent module scope for the logger
            # so this does not hit (and get handled by) the local
            # pyomo.common.timing logger.
            deprecation_warning(
                "toc(): 'delta', 'ostream', and 'logger' should be "
                "specified as keyword arguments",
                version='6.4.2',
                logger=__package__,
            )
            delta, *args = args
            if args:
                ostream, *args = args
            if args:
                logger, *args = args

        if self._start_count or self._lastTime is None:
            ans = self._cumul
            if self._lastTime:
                ans += now - self._lastTime
            if msg is not None:
                fmt = "[%8.2f|%4d] %s"
                data = (ans, self._start_count, msg)
        elif delta:
            ans = now - self._lastTime
            self._lastTime = now
            if msg is not None:
                fmt = "[+%7.2f] %s"
                data = (ans, msg)
        else:
            ans = now - self._loadTime
            # Even though we are reporting the cumulative time, we will
            # still reset the delta timer.
            self._lastTime = now
            if msg is not None:
                fmt = "[%8.2f] %s"
                data = (ans, msg)

        if msg is not None:
            if logger is _NotSpecified:
                logger = self.logger
            if logger is not None:
                if level is _NotSpecified:
                    level = self.level
                logger.log(level, GeneralTimer(fmt, data), *args)

            if ostream is _NotSpecified:
                ostream = self.ostream
                if ostream is _NotSpecified:
                    if logger is None:
                        ostream = sys.stdout
                    else:
                        ostream = None
            if ostream is not None:
                msg = fmt % data
                if args:
                    msg = msg % args
                ostream.write(msg + '\n')

        return ans

    def stop(self):
        delta, self._lastTime = self._lastTime, None
        if delta is None:
            raise RuntimeError("Stopping a TicTocTimer that was already stopped")
        delta = default_timer() - delta
        self._cumul += delta
        return delta

    def start(self):
        if self._lastTime:
            self.stop()
        self._start_count += 1
        self._lastTime = default_timer()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, et, ev, tb):
        self.stop()


_globalTimer = TicTocTimer()
tic = functools.partial(TicTocTimer.tic, _globalTimer)
tic.__doc__ = """
Reset the global :py:class:`TicTocTimer` instance.

See :py:meth:`TicTocTimer.tic()`.
"""
toc = functools.partial(TicTocTimer.toc, _globalTimer)
toc.__doc__ = """
Print the elapsed time from the global :py:class:`TicTocTimer` instance.

See :py:meth:`TicTocTimer.toc()`.
"""


def _move_grandchildren_to_root(root, child):
    """A helper function to assist with flattening of HierarchicalTimer
    objects

    Parameters
    ----------
    root: HierarchicalTimer or _HierarchicalHelper
        The root node. Children of `child` will become children of
        this node

    child: _HierarchicalHelper
        The child node that will be turned into a leaf by moving
        its children to the root

    """
    for gchild_key, gchild_timer in child.timers.items():
        # For each grandchild, if this key corresponds to a child,
        # combine the information from these timers. Otherwise,
        # add the new timer as a child of the root.
        if gchild_key in root.timers:
            gchild_total_time = gchild_timer.total_time
            gchild_n_calls = gchild_timer.n_calls
            root.timers[gchild_key].total_time += gchild_total_time
            root.timers[gchild_key].n_calls += gchild_n_calls
        else:
            root.timers[gchild_key] = gchild_timer

        # Subtract the grandchild's total time from the child (which
        # will no longer be a parent of the grandchild)
        child.total_time -= gchild_timer.total_time

    # Clear the child timer's dict to make it a leaf node
    child.timers.clear()


def _clear_timers_except(timer, to_retain):
    """A helper function for removing keys, except for those specified,
    from the dictionary of timers

    Parameters
    ----------
    timer: HierarchicalTimer or _HierarchicalHelper
        The timer whose dict of "sub-timers" will be pruned

    to_retain: set
        Set of keys of the "sub-timers" to retain

    """
    keys = list(timer.timers.keys())
    for key in keys:
        if key not in to_retain:
            timer.timers.pop(key)


class _HierarchicalHelper:
    def __init__(self):
        self.tic_toc = TicTocTimer()
        self.timers = dict()
        self.total_time = 0
        self.n_calls = 0

    def start(self):
        self.n_calls += 1
        self.tic_toc.start()

    def stop(self):
        self.total_time += self.tic_toc.stop()

    def to_str(self, indent, stage_identifier_lengths):
        s = ''
        if len(self.timers) > 0:
            underline = indent + '-' * (sum(stage_identifier_lengths) + 36) + '\n'
            s += underline
            name_formatter = '{name:<' + str(sum(stage_identifier_lengths)) + '}'
            other_time = self.total_time
            sub_stage_identifier_lengths = stage_identifier_lengths[1:]
            for name, timer in sorted(self.timers.items()):
                if self.total_time > 0:
                    _percent = timer.total_time / self.total_time * 100
                else:
                    _percent = float('nan')
                s += indent
                s += (
                    name_formatter + '{ncalls:>9d} {cumtime:>9.3f} '
                    '{percall:>9.3f} {percent:>6.1f}\n'
                ).format(
                    name=name,
                    ncalls=timer.n_calls,
                    cumtime=timer.total_time,
                    percall=timer.total_time / timer.n_calls,
                    percent=_percent,
                )
                s += timer.to_str(
                    indent=indent + ' ' * stage_identifier_lengths[0],
                    stage_identifier_lengths=sub_stage_identifier_lengths,
                )
                other_time -= timer.total_time

            if self.total_time > 0:
                _percent = other_time / self.total_time * 100
            else:
                _percent = float('nan')
            s += indent
            s += (
                name_formatter + '{ncalls:>9} {cumtime:>9.3f} '
                '{percall:>9} {percent:>6.1f}\n'
            ).format(
                name='other',
                ncalls='n/a',
                cumtime=other_time,
                percall='n/a',
                percent=_percent,
            )
            s += underline.replace('-', '=')
        return s

    def get_timers(self, res, prefix):
        for name, timer in self.timers.items():
            _name = prefix + '.' + name
            res.append(_name)
            timer.get_timers(res, _name)

    def flatten(self):
        # Get keys and values so we don't modify dict while iterating it.
        items = list(self.timers.items())
        for child_key, child_timer in items:
            # Flatten the child timer. Now all grandchildren are leaf nodes
            child_timer.flatten()
            # Flatten by removing grandchildren and adding them as children
            # of the root.
            _move_grandchildren_to_root(self, child_timer)

    def clear_except(self, *args):
        to_retain = set(args)
        _clear_timers_except(self, to_retain)


class HierarchicalTimer:
    """A class for collecting and displaying hierarchical timing
    information

    When implementing an iterative algorithm with nested subroutines
    (e.g. an optimization solver), we often want to know the cumulative
    time spent in each subroutine as well as this time as a proportion
    of time spent in the calling routine. This class collects timing
    information, for user-specified keys, that accumulates over the life
    of the timer object and preserves the hierarchical (nested) structure
    of timing categories.

    Examples
    --------
    >>> import time
    >>> from pyomo.common.timing import HierarchicalTimer
    >>> timer = HierarchicalTimer()
    >>> timer.start('all')
    >>> time.sleep(0.2)
    >>> for i in range(10):
    ...     timer.start('a')
    ...     time.sleep(0.1)
    ...     for i in range(5):
    ...         timer.start('aa')
    ...         time.sleep(0.01)
    ...         timer.stop('aa')
    ...     timer.start('ab')
    ...     timer.stop('ab')
    ...     timer.stop('a')
    ...
    >>> for i in range(10):
    ...     timer.start('b')
    ...     time.sleep(0.02)
    ...     timer.stop('b')
    ...
    >>> timer.stop('all')
    >>> print(timer)       # doctest: +SKIP
    Identifier        ncalls   cumtime   percall      %
    ---------------------------------------------------
    all                    1     2.248     2.248  100.0
         ----------------------------------------------
         a                10     1.787     0.179   79.5
              -----------------------------------------
              aa          50     0.733     0.015   41.0
              ab          10     0.000     0.000    0.0
              other      n/a     1.055       n/a   59.0
              =========================================
         b                10     0.248     0.025   11.0
         other           n/a     0.213       n/a    9.5
         ==============================================
    ===================================================
    <BLANKLINE>

    The columns are:

      ncalls
          The number of times the timer was started and stopped
      cumtime
          The cumulative time (in seconds) the timer was active
          (started but not stopped)
      percall
          cumtime (in seconds) / ncalls
      "%"
          This is cumtime of the timer divided by cumtime of the
          parent timer times 100

    >>> print('a total time: %f' % timer.get_total_time('all.a')) \
        # doctest: +SKIP
    a total time: 1.902037
    >>> print('ab num calls: %d' % timer.get_num_calls('all.a.ab')) \
        # doctest: +SKIP
    ab num calls: 10
    >>> print('aa %% time: %f' % timer.get_relative_percent_time('all.a.aa')) \
        # doctest: +SKIP
    aa % time: 44.144148
    >>> print('aa %% total: %f' % timer.get_total_percent_time('all.a.aa')) \
        # doctest: +SKIP
    aa % total: 35.976058

    When implementing an algorithm, it is often useful to collect detailed
    hierarchical timing information. However, when communicating a timing
    profile, it is often best to retain only the most relevant information
    in a flattened data structure. In the following example, suppose we
    want to compare the time spent in the ``"c"`` and ``"f"`` subroutines.
    We would like to generate a timing profile that displays only the time
    spent in these two subroutines, in a flattened structure so that they
    are easy to compare. To do this, we

    #. Ignore subroutines of ``"c"`` and ``"f"`` that are unnecessary for\
    this comparison

    #. Flatten the hierarchical timing information

    #. Eliminate all the information we don't care about

    >>> import time
    >>> from pyomo.common.timing import HierarchicalTimer
    >>> timer = HierarchicalTimer()
    >>> timer.start("root")
    >>> timer.start("a")
    >>> time.sleep(0.01)
    >>> timer.start("b")
    >>> timer.start("c")
    >>> time.sleep(0.1)
    >>> timer.stop("c")
    >>> timer.stop("b")
    >>> timer.stop("a")
    >>> timer.start("d")
    >>> timer.start("e")
    >>> time.sleep(0.01)
    >>> timer.start("f")
    >>> time.sleep(0.05)
    >>> timer.stop("f")
    >>> timer.start("c")
    >>> timer.start("g")
    >>> timer.start("h")
    >>> time.sleep(0.1)
    >>> timer.stop("h")
    >>> timer.stop("g")
    >>> timer.stop("c")
    >>> timer.stop("e")
    >>> timer.stop("d")
    >>> timer.stop("root")
    >>> print(timer) # doctest: +SKIP
    Identifier                       ncalls   cumtime   percall      %
    ------------------------------------------------------------------
    root                                  1     0.290     0.290  100.0
         -------------------------------------------------------------
         a                                1     0.118     0.118   40.5
              --------------------------------------------------------
              b                           1     0.105     0.105   89.4
                   ---------------------------------------------------
                   c                      1     0.105     0.105  100.0
                   other                n/a     0.000       n/a    0.0
                   ===================================================
              other                     n/a     0.013       n/a   10.6
              ========================================================
         d                                1     0.173     0.173   59.5
              --------------------------------------------------------
              e                           1     0.173     0.173  100.0
                   ---------------------------------------------------
                   c                      1     0.105     0.105   60.9
                        ----------------------------------------------
                        g                 1     0.105     0.105  100.0
                             -----------------------------------------
                             h            1     0.105     0.105  100.0
                             other      n/a     0.000       n/a    0.0
                             =========================================
                        other           n/a     0.000       n/a    0.0
                        ==============================================
                   f                      1     0.055     0.055   31.9
                   other                n/a     0.013       n/a    7.3
                   ===================================================
              other                     n/a     0.000       n/a    0.0
              ========================================================
         other                          n/a     0.000       n/a    0.0
         =============================================================
    ==================================================================
    >>> # Clear subroutines under "c" that we don't care about
    >>> timer.timers["root"].timers["d"].timers["e"].timers["c"].timers.clear()
    >>> # Flatten hierarchy
    >>> timer.timers["root"].flatten()
    >>> # Clear except for the subroutines we care about
    >>> timer.timers["root"].clear_except("c", "f")
    >>> print(timer) # doctest: +SKIP
    Identifier   ncalls   cumtime   percall      %
    ----------------------------------------------
    root              1     0.290     0.290  100.0
         -----------------------------------------
         c            2     0.210     0.105   72.4
         f            1     0.055     0.055   19.0
         other      n/a     0.025       n/a    8.7
         =========================================
    ==============================================

    Notes
    -----

    The :py:class:`HierarchicalTimer` uses a stack to track which timers
    are active at any point in time. Additionally, each timer has a
    dictionary of timers for its children timers. Consider

    >>> timer = HierarchicalTimer()
    >>> timer.start('all')
    >>> timer.start('a')
    >>> timer.start('aa')

    After the above code is run, ``timer.stack`` will be
    ``['all', 'a', 'aa']`` and ``timer.timers`` will have one key,
    ``'all'`` and one value which will be a
    :py:class:`_HierarchicalHelper`. The :py:class:`_HierarchicalHelper`
    has its own timers dictionary:

        ``{'a': _HierarchicalHelper}``

    and so on. This way, we can easily access any timer with something
    that looks like the stack. The logic is recursive (although the
    code is not).

    """

    def __init__(self):
        self.stack = list()
        self.timers = dict()

    def _get_timer(self, identifier, should_exist=False):
        """
        This method gets the timer associated with the current state
        of self.stack and the specified identifier.

        Parameters
        ----------
        identifier: str
            The name of the timer
        should_exist: bool
            The should_exist is True, and the timer does not already
            exist, an error will be raised. If should_exist is False, and
            the timer does not already exist, a new timer will be made.

        Returns
        -------
        timer: _HierarchicalHelper

        """
        parent = self._get_timer_from_stack(self.stack)
        if identifier in parent.timers:
            return parent.timers[identifier]
        else:
            if should_exist:
                raise RuntimeError(
                    'Could not find timer {0}'.format(
                        '.'.join(self.stack + [identifier])
                    )
                )
            parent.timers[identifier] = _HierarchicalHelper()
            return parent.timers[identifier]

    def start(self, identifier):
        """Start incrementing the timer identified with identifier

        Parameters
        ----------
        identifier: str
            The name of the timer
        """
        timer = self._get_timer(identifier)
        timer.start()
        self.stack.append(identifier)

    def stop(self, identifier):
        """Stop incrementing the timer identified with identifier

        Parameters
        ----------
        identifier: str
            The name of the timer
        """
        if identifier != self.stack[-1]:
            raise ValueError(
                str(identifier) + ' is not the currently active timer.  '
                'The only timer that can currently be stopped is '
                + '.'.join(self.stack)
            )
        self.stack.pop()
        timer = self._get_timer(identifier, should_exist=True)
        timer.stop()

    def _get_identifier_len(self):
        stage_timers = list(self.timers.items())
        stage_lengths = list()

        while len(stage_timers) > 0:
            new_stage_timers = list()
            max_len = 0
            for identifier, timer in stage_timers:
                new_stage_timers.extend(timer.timers.items())
                if len(identifier) > max_len:
                    max_len = len(identifier)
            stage_lengths.append(max(max_len, len('other')))
            stage_timers = new_stage_timers

        return stage_lengths

    def __str__(self):
        const_indent = 4
        max_name_length = 200 - 36
        stage_identifier_lengths = self._get_identifier_len()
        name_field_width = sum(stage_identifier_lengths)
        if name_field_width > max_name_length:
            # switch to a constant indentation of const_indent spaces
            # (to hopefully shorten the line lengths
            name_field_width = max(
                const_indent * i + l for i, l in enumerate(stage_identifier_lengths)
            )
            for i in range(len(stage_identifier_lengths) - 1):
                stage_identifier_lengths[i] = const_indent
            stage_identifier_lengths[-1] = name_field_width - const_indent * (
                len(stage_identifier_lengths) - 1
            )
        name_formatter = '{name:<' + str(name_field_width) + '}'
        s = (
            name_formatter + '{ncalls:>9} {cumtime:>9} {percall:>9} {percent:>6}\n'
        ).format(
            name='Identifier',
            ncalls='ncalls',
            cumtime='cumtime',
            percall='percall',
            percent='%',
        )
        underline = '-' * (name_field_width + 36) + '\n'
        s += underline
        sub_stage_identifier_lengths = stage_identifier_lengths[1:]
        for name, timer in sorted(self.timers.items()):
            s += (
                name_formatter + '{ncalls:>9d} {cumtime:>9.3f} '
                '{percall:>9.3f} {percent:>6.1f}\n'
            ).format(
                name=name,
                ncalls=timer.n_calls,
                cumtime=timer.total_time,
                percall=timer.total_time / timer.n_calls,
                percent=self.get_total_percent_time(name),
            )
            s += timer.to_str(
                indent=' ' * stage_identifier_lengths[0],
                stage_identifier_lengths=sub_stage_identifier_lengths,
            )
        s += underline.replace('-', '=')
        return s

    def reset(self):
        """
        Completely reset the timer.
        """
        self.stack = list()
        self.timers = dict()

    def _get_timer_from_stack(self, stack):
        """
        This method gets the timer associated with stack.

        Parameters
        ----------
        stack: list of str
            A list of identifiers.

        Returns
        -------
        timer: _HierarchicalHelper
        """
        tmp = self
        for i in stack:
            tmp = tmp.timers[i]
        return tmp

    def get_total_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        total_time: float
            The total time spent with the specified timer active.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        return timer.total_time

    def get_num_calls(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        n_calls: int
            The number of times start was called for the specified timer.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        return timer.n_calls

    def get_relative_percent_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        percent_time: float
            The percent of time spent in the specified timer
            relative to the timer's immediate parent.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        parent = self._get_timer_from_stack(stack[:-1])
        if parent is self:
            return self.get_total_percent_time(identifier)
        else:
            if parent.total_time > 0:
                return timer.total_time / parent.total_time * 100
            else:
                return float('nan')

    def get_total_percent_time(self, identifier):
        """
        Parameters
        ----------
        identifier: str
            The full name of the timer including parent timers separated
            with dots.

        Returns
        -------
        percent_time: float
            The percent of time spent in the specified timer
            relative to the total time in all timers.
        """
        stack = identifier.split('.')
        timer = self._get_timer_from_stack(stack)
        total_time = 0
        for _timer in self.timers.values():
            total_time += _timer.total_time
        if total_time > 0:
            return timer.total_time / total_time * 100
        else:
            return float('nan')

    def get_timers(self):
        """
        Returns
        -------
        identifiers: list of str
            Returns a list of all timer identifiers
        """
        res = list()
        for name, timer in self.timers.items():
            res.append(name)
            timer.get_timers(res, name)
        return res

    def flatten(self):
        """Flatten the HierarchicalTimer in-place, moving all the timing
        categories into a single level

        If any timers moved into the same level have the same identifier,
        the ``total_time`` and ``n_calls`` fields are added together.
        The ``total_time`` of a "child timer" that is "moved upwards" is
        subtracted from the ``total_time`` of that timer's original
        parent.

        """
        if self.stack:
            raise RuntimeError(
                "Cannot flatten a HierarchicalTimer while any timers are"
                " active. Current active timer is %s. flatten should only"
                " be called as a post-processing step." % self.stack[-1]
            )
        items = list(self.timers.items())
        for key, timer in items:
            timer.flatten()
            _move_grandchildren_to_root(self, timer)

    def clear_except(self, *args):
        """Prune all "sub-timers" except those specified

        Parameters
        ----------
        args: str
            Keys that will be retained

        """
        if self.stack:
            raise RuntimeError(
                "Cannot clear a HierarchicalTimer while any timers are"
                " active. Current active timer is %s. clear_except should"
                " only be called as a post-processing step." % self.stack[-1]
            )
        to_retain = set(args)
        _clear_timers_except(self, to_retain)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Utility functions
#

from pyomo.common.deprecation import deprecation_warning
from pyomo.core.expr.numvalue import native_numeric_types
from pyomo.core.expr.numeric_expr import mutable_expression, NPV_SumExpression
from pyomo.core.base.var import Var
from pyomo.core.base.expression import Expression
from pyomo.core.base.component import ComponentBase
import logging

logger = logging.getLogger(__name__)


def prod(terms):
    """
    A utility function to compute the product of a list of terms.

    Args:
        terms (list): A list of terms that are multiplied together.

    Returns:
        The value of the product, which may be a Pyomo expression object.
    """
    ans = 1
    for term in terms:
        ans *= term
    return ans


def quicksum(args, start=0, linear=None):
    """A utility function to compute a sum of Pyomo expressions.

    The behavior of :func:`quicksum` is similar to the builtin
    :func:`sum` function, but this function can avoid the generation and
    disposal of intermediate objects, and thus is slightly more
    performant.

    Parameters
    ----------
    args: Iterable
        A generator for terms in the sum.

    start: Any
        A value that initializes the sum.  If this value is not a
        numeric constant, then the += operator is used to add terms to
        this object.  Defaults to 0.

    linear: bool
        DEPRECATED: the linearity of the resulting expression is
        determined automatically.  This option is ignored.

    Returns
    -------
    The value of the sum, which may be a Pyomo expression object.

    """

    # Ensure that args is an iterator (this manages things like
    # IndexedComponent_slice objects)
    try:
        args = iter(args)
    except:
        logger.error('The argument `args` to quicksum() is not iterable!')
        raise

    if linear is not None:
        deprecation_warning(
            "The quicksum(linear=...) argument is deprecated and ignored.",
            version='6.6.0',
        )

    #
    # If we're starting with a numeric value, then
    # create a new nonlinear sum expression but
    # return a static version to the user.
    #
    if start.__class__ in native_numeric_types:
        with mutable_expression() as e:
            e += start
            for arg in args:
                e += arg
        # Special case: reduce NPV sums of native types to a single
        # constant
        if e.__class__ is NPV_SumExpression and all(
            arg.__class__ in native_numeric_types for arg in e.args
        ):
            return e()
        if e.nargs() > 1:
            return e
        elif not e.nargs():
            return 0
        else:
            return e.arg(0)
    #
    # Otherwise, use the context that is provided and return it.
    #
    e = start
    for arg in args:
        e += arg
    return e


def sum_product(*args, **kwds):
    """
    A utility function to compute a generalized dot product.

    This function accepts one or more components that provide terms
    that are multiplied together.  These products are added together
    to form a sum.

    Args:
        *args: Variable length argument list of generators that
            create terms in the summation.
        **kwds: Arbitrary keyword arguments.

    Keyword Args:
        index: A set that is used to index the components used to
            create the terms
        denom: A component or tuple of components that are used to
            create the denominator of the terms
        start: The initial value used in the sum

    Returns:
        The value of the sum.
    """
    denom = kwds.pop('denom', tuple())
    if type(denom) not in (list, tuple):
        denom = [denom]
    nargs = len(args)
    ndenom = len(denom)

    if nargs == 0 and ndenom == 0:
        raise ValueError(
            "The sum_product() command requires at least an "
            + "argument or a denominator term"
        )

    if 'index' in kwds:
        index = kwds['index']
    else:
        if nargs > 0:
            iarg = args[-1]
            if not isinstance(iarg, Var) and not isinstance(iarg, Expression):
                raise ValueError(
                    "Error executing sum_product(): The last argument value must be a variable or expression object if no 'index' option is specified"
                )
        else:
            iarg = denom[-1]
            if not isinstance(iarg, Var) and not isinstance(iarg, Expression):
                raise ValueError(
                    "Error executing sum_product(): The last denom argument value must be a variable or expression object if no 'index' option is specified"
                )
        index = iarg.index_set()

    start = kwds.get("start", 0)

    if ndenom == 0:
        #
        # Sum of polynomial terms
        #
        with mutable_expression() as expr:
            expr += start
            if nargs == 1:
                arg1 = args[0]
                for i in index:
                    expr += arg1[i]
            elif nargs == 2:
                arg1, arg2 = args
                for i in index:
                    expr += arg1[i] * arg2[i]
            else:
                for i in index:
                    expr += prod(arg[i] for arg in args)
        if expr.nargs() > 1:
            return expr
        elif not expr.nargs():
            return 0
        else:
            return expr.arg(0)
    elif nargs == 0:
        #
        # Sum of reciprocals
        #
        return quicksum((1 / prod(den[i] for den in denom) for i in index), start)
    else:
        #
        # Sum of fractions
        #
        return quicksum(
            (
                prod(arg[i] for arg in args) / prod(den[i] for den in denom)
                for i in index
            ),
            start,
        )


#: An alias for :func:`sum_product <pyomo.core.expr.util>`
dot_product = sum_product

#: An alias for :func:`sum_product <pyomo.core.expr.util>`
summation = sum_product


def sequence(*args):
    """
    sequence([start,] stop[, step]) -> generator for a list of integers

    Return a generator that containing an arithmetic
    progression of integers.

       - ``sequence(i, j)`` returns ``[i, i+1, i+2, ..., j]``;
       - start defaults to 1.
       - step specifies the increment (or decrement)

    For example, ``sequence(4)`` returns ``[1, 2, 3, 4]``.
    """
    if len(args) == 0:
        raise ValueError('sequence expected at least 1 arguments, got 0')
    if len(args) > 3:
        raise ValueError('sequence expected at most 3 arguments, got %d' % len(args))
    if len(args) == 1:
        return range(1, args[0] + 1)
    if len(args) == 2:
        return range(args[0], args[1] + 1)
    return range(args[0], args[1] + 1, args[2])


def target_list(x):
    if isinstance(x, ComponentBase):
        return [x]
    elif hasattr(x, '__iter__'):
        ans = []
        for i in x:
            if isinstance(i, ComponentBase):
                ans.append(i)
            else:
                raise ValueError(
                    "Expected Component or list of Components."
                    "\n\tReceived %s" % (type(i),)
                )
        return ans
    else:
        raise ValueError(
            "Expected Component or list of Components.\n\tReceived %s" % (type(x),)
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pympler_available
from pyomo.common.collections import Bunch
import pyomo.scripting.util
from pyomo.core import ConcreteModel


def run_pyomo(options=Bunch(), parser=None):
    data = Bunch(options=options)

    if options.model.filename == '':
        parser.print_help()
        return Bunch()

    try:
        pyomo.scripting.util.setup_environment(data)

        pyomo.scripting.util.apply_preprocessing(data, parser=parser)
    except:
        # TBD: I should be able to call this function in the case of
        #      an exception to perform cleanup. However, as it stands
        #      calling finalize with its default keyword value for
        #      model(=None) results in an a different error related to
        #      task port values.  Not sure how to interpret that.
        pyomo.scripting.util.finalize(
            data, model=ConcreteModel(), instance=None, results=None
        )
        raise
    else:
        if data.error:
            # TBD: I should be able to call this function in the case of
            #      an exception to perform cleanup. However, as it stands
            #      calling finalize with its default keyword value for
            #      model(=None) results in an a different error related to
            #      task port values.  Not sure how to interpret that.
            pyomo.scripting.util.finalize(
                data, model=ConcreteModel(), instance=None, results=None
            )
            return Bunch()  # pragma:nocover

    try:
        model_data = pyomo.scripting.util.create_model(data)
    except:
        # TBD: I should be able to call this function in the case of
        #      an exception to perform cleanup. However, as it stands
        #      calling finalize with its default keyword value for
        #      model(=None) results in an a different error related to
        #      task port values.  Not sure how to interpret that.
        pyomo.scripting.util.finalize(
            data, model=ConcreteModel(), instance=None, results=None
        )
        raise
    else:
        if (
            (not options.runtime.logging == 'debug') and options.model.save_file
        ) or options.runtime.only_instance:
            pyomo.scripting.util.finalize(
                data, model=model_data.model, instance=model_data.instance, results=None
            )
            return Bunch(instance=model_data.instance)

    try:
        opt_data = pyomo.scripting.util.apply_optimizer(
            data, instance=model_data.instance
        )

        pyomo.scripting.util.process_results(
            data,
            instance=model_data.instance,
            results=opt_data.results,
            opt=opt_data.opt,
        )

        pyomo.scripting.util.apply_postprocessing(
            data, instance=model_data.instance, results=opt_data.results
        )
    except:
        # TBD: I should be able to call this function in the case of
        #      an exception to perform cleanup. However, as it stands
        #      calling finalize with its default keyword value for
        #      model(=None) results in an a different error related to
        #      task port values.  Not sure how to interpret that.
        pyomo.scripting.util.finalize(
            data, model=ConcreteModel(), instance=None, results=None
        )
        raise
    else:
        pyomo.scripting.util.finalize(
            data,
            model=model_data.model,
            instance=model_data.instance,
            results=opt_data.results,
        )

        return Bunch(
            options=options,
            instance=model_data.instance,
            results=opt_data.results,
            local=opt_data.local,
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import inspect
import sys
from importlib.machinery import PathFinder
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.log import LoggingIntercept
from pyomo.common.dependencies import (
    attempt_import,
    ModuleUnavailable,
    DeferredImportModule,
    DeferredImportIndicator,
    DeferredImportError,
    UnavailableClass,
    _DeferredAnd,
    _DeferredOr,
    _DeferredImportCallbackFinder,
    check_min_version,
    dill,
    dill_available,
    mpi4py_available,
    packaging_available,
)

import pyomo.common.tests.dep_mod as dep_mod

from . import deps


# global objects for the submodule tests
def _finalize_pyo(module, available):
    if available:
        import pyomo.core


class TestDependencies(unittest.TestCase):
    def test_import_error(self):
        module_obj, module_available = attempt_import(
            '__there_is_no_module_named_this__',
            'Testing import of a non-existent module',
            defer_import=False,
        )
        self.assertFalse(module_available)
        with self.assertRaisesRegex(
            DeferredImportError, 'Testing import of a non-existent module'
        ):
            module_obj.try_to_call_a_method()

        # Note that some attribute will intentionally raise
        # AttributeErrors and NOT DeferredImportError:
        with self.assertRaisesRegex(
            AttributeError,
            "'ModuleUnavailable' object has no attribute '__sphinx_mock__'",
        ):
            module_obj.__sphinx_mock__

    @unittest.skipUnless(dill_available, "Test requires dill module")
    def test_pickle(self):
        self.assertIs(deps.pkl_test.__class__, DeferredImportModule)
        # Pickle the DeferredImportModule class
        pkl = dill.dumps(deps.pkl_test)
        deps.new_pkl_test = dill.loads(pkl)
        self.assertIs(deps.pkl_test.__class__, deps.new_pkl_test.__class__)
        self.assertIs(deps.new_pkl_test.__class__, DeferredImportModule)
        self.assertIsNot(deps.pkl_test, deps.new_pkl_test)
        self.assertIn('submod', deps.new_pkl_test.__dict__)
        with self.assertRaisesRegex(
            DeferredImportError, 'nonexisting.module.pickle_test module'
        ):
            deps.new_pkl_test.try_to_call_a_method()
        # Pickle the ModuleUnavailable class
        self.assertIs(deps.new_pkl_test.__class__, ModuleUnavailable)
        pkl = dill.dumps(deps.new_pkl_test)
        new_pkl_test_2 = dill.loads(pkl)
        self.assertIs(deps.new_pkl_test.__class__, new_pkl_test_2.__class__)
        self.assertIsNot(deps.new_pkl_test, new_pkl_test_2)
        self.assertIs(new_pkl_test_2.__class__, ModuleUnavailable)

    def test_import_success(self):
        module_obj, module_available = attempt_import(
            'pyomo.tpl.ply', 'Testing import of ply', defer_import=False
        )
        self.assertTrue(module_available)
        import pyomo.tpl.ply

        self.assertTrue(module_obj is pyomo.tpl.ply)

    def test_local_deferred_import(self):
        self.assertIs(type(deps.bogus_available), DeferredImportIndicator)
        self.assertIs(type(deps.bogus), DeferredImportModule)
        if deps.bogus_available:
            self.fail("Casting bogus_available to bool returned True")
        self.assertIs(deps.bogus_available, False)
        # Note: this also tests the implicit alt_names for dotted imports
        self.assertIs(type(deps.bogus), ModuleUnavailable)
        with self.assertRaisesRegex(
            DeferredImportError,
            "The nonexisting.module.bogus module "
            r"\(an optional Pyomo dependency\) failed to import",
        ):
            deps.bogus.hello

    def test_imported_deferred_import(self):
        self.assertIs(type(deps.has_bogus_nem), DeferredImportIndicator)
        self.assertIs(type(deps.bogus_nem), DeferredImportModule)
        with self.assertRaisesRegex(
            DeferredImportError,
            "The bogus_nonexisting_module module "
            r"\(an optional Pyomo dependency\) failed to import",
        ):
            deps.test_access_bogus_hello()
        self.assertIs(deps.has_bogus_nem, False)
        self.assertIs(type(deps.bogus_nem), ModuleUnavailable)
        self.assertIs(dep_mod.bogus_nonexisting_module_available, False)
        self.assertIs(type(dep_mod.bogus_nonexisting_module), ModuleUnavailable)

    @unittest.skipUnless(
        packaging_available, "min_version tests require packaging module"
    )
    def test_min_version(self):
        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod', minimum_version='1.0', defer_import=False
        )
        self.assertTrue(avail)
        self.assertTrue(inspect.ismodule(mod))
        self.assertTrue(check_min_version(mod, '1.0'))
        self.assertFalse(check_min_version(mod, '2.0'))

        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod', minimum_version='2.0', defer_import=False
        )
        self.assertFalse(avail)
        self.assertIs(type(mod), ModuleUnavailable)
        with self.assertRaisesRegex(
            DeferredImportError,
            "The pyomo.common.tests.dep_mod module "
            "version 1.5 does not satisfy the minimum version 2.0",
        ):
            mod.hello

        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod',
            error_message="Failed import",
            minimum_version='2.0',
            defer_import=False,
        )
        self.assertFalse(avail)
        self.assertIs(type(mod), ModuleUnavailable)
        with self.assertRaisesRegex(
            DeferredImportError,
            "Failed import "
            r"\(version 1.5 does not satisfy the minimum version 2.0\)",
        ):
            mod.hello

        # Verify check_min_version works with deferred imports

        mod, avail = attempt_import('pyomo.common.tests.dep_mod', defer_import=True)
        self.assertTrue(check_min_version(mod, '1.0'))

        mod, avail = attempt_import('pyomo.common.tests.dep_mod', defer_import=True)
        self.assertFalse(check_min_version(mod, '2.0'))

        # Verify check_min_version works when called directly

        mod, avail = attempt_import('pyomo.common.tests.dep_mod', minimum_version='1.0')
        self.assertTrue(check_min_version(mod, '1.0'))

        mod, avail = attempt_import('pyomo.common.tests.bogus', minimum_version='1.0')
        self.assertFalse(check_min_version(mod, '1.0'))

    @unittest.skipUnless(
        packaging_available, "min_version tests require packaging module"
    )
    def test_and_or(self):
        mod0, avail0 = attempt_import('pyomo.tpl.ply', defer_import=True)
        mod1, avail1 = attempt_import('pyomo.common.tests.dep_mod', defer_import=True)
        mod2, avail2 = attempt_import(
            'pyomo.common.tests.dep_mod', minimum_version='2.0', defer_import=True
        )

        _and = avail0 & avail1
        self.assertIsInstance(_and, _DeferredAnd)

        _or = avail1 | avail2
        self.assertIsInstance(_or, _DeferredOr)

        # Nothing has been resolved yet
        self.assertIsNone(avail0._available)
        self.assertIsNone(avail1._available)
        self.assertIsNone(avail2._available)

        # Shortcut boolean evaluation only partially resolves things
        self.assertTrue(_or)
        self.assertIsNone(avail0._available)
        self.assertTrue(avail1._available)
        self.assertIsNone(avail2._available)

        self.assertTrue(_and)
        self.assertTrue(avail0._available)
        self.assertTrue(avail1._available)
        self.assertIsNone(avail2._available)

        # Testing compound operations
        _and_and = avail0 & avail1 & avail2
        self.assertFalse(_and_and)

        _and_or = avail0 & avail1 | avail2
        self.assertTrue(_and_or)

        # Verify operator precedence
        _or_and = avail0 | avail2 & avail2
        self.assertTrue(_or_and)
        _or_and = (avail0 | avail2) & avail2
        self.assertFalse(_or_and)

        _or_or = avail0 | avail1 | avail2
        self.assertTrue(_or_or)

        # Verify rand / ror
        _rand = True & avail1
        self.assertIsInstance(_rand, _DeferredAnd)
        self.assertTrue(_rand)

        _ror = False | avail1
        self.assertIsInstance(_ror, _DeferredOr)
        self.assertTrue(_ror)

    @unittest.skipUnless(
        packaging_available, "min_version tests require packaging module"
    )
    def test_callbacks(self):
        ans = []

        def _record_avail(module, avail):
            ans.append(avail)

        mod0, avail0 = attempt_import(
            'pyomo.tpl.ply', defer_import=True, callback=_record_avail
        )
        mod1, avail1 = attempt_import(
            'pyomo.common.tests.dep_mod',
            minimum_version='2.0',
            defer_import=True,
            callback=_record_avail,
        )

        self.assertEqual(ans, [])
        self.assertTrue(avail0)
        self.assertEqual(ans, [True])
        self.assertFalse(avail1)
        self.assertEqual(ans, [True, False])

    def test_callback_on_import(self):
        sys.modules.pop('pyomo.common.tests.mod', None)
        ans = []

        class ImpFinder:
            # This is an "imp" module-style finder (deprecated in Python
            # 3.4 and removed in Python 3.12, but Google Collab still
            # defines finders like this)
            match = ''

            def find_module(self, fullname, path=None):
                if fullname != self.match:
                    ans.append('pass')
                    return None
                ans.append('load')
                spec = PathFinder().find_spec(fullname, path)
                return spec.loader

            def load_module(self, name):
                pass

        def _callback(module, avail):
            ans.append(len(ans))

        attempt_import('pyomo.common.tests.mod', defer_import=True, callback=_callback)
        self.assertEqual(ans, [])
        import pyomo.common.tests.mod as m

        self.assertEqual(ans, [0])
        self.assertEqual(m.Foo.data, 42)

        sys.modules.pop('pyomo.common.tests.mod', None)
        del m
        attempt_import('pyomo.common.tests.mod', defer_import=True, callback=_callback)

        try:
            # Test deferring to an imp-style finder that does not match
            # the target module name
            _finder = ImpFinder()
            sys.meta_path.insert(
                sys.meta_path.index(_DeferredImportCallbackFinder) + 1, _finder
            )
            import pyomo.common.tests.mod as m

            self.assertEqual(ans, [0, 'pass', 2])
            self.assertEqual(m.Foo.data, 42)

            sys.modules.pop('pyomo.common.tests.mod', None)
            del m
            attempt_import(
                'pyomo.common.tests.mod', defer_import=True, callback=_callback
            )

            # Test deferring to an imp-style finder that DOES match the
            # target module name
            _finder.match = 'pyomo.common.tests.mod'

            import pyomo.common.tests.mod as m

            self.assertEqual(ans, [0, 'pass', 2, 'load', 4])
            self.assertEqual(m.Foo.data, 42)
        finally:
            sys.meta_path.remove(_finder)

    def test_import_exceptions(self):
        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod_except',
            defer_import=True,
            only_catch_importerror=True,
        )
        with self.assertRaisesRegex(ValueError, "cannot import module"):
            bool(avail)
        # second test will not re-trigger the exception
        self.assertFalse(avail)

        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod_except',
            defer_import=True,
            only_catch_importerror=False,
        )
        self.assertFalse(avail)
        self.assertFalse(avail)

        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod_except',
            defer_import=True,
            catch_exceptions=(ImportError, ValueError),
        )
        self.assertFalse(avail)
        self.assertFalse(avail)

        with self.assertRaisesRegex(
            ValueError,
            'Cannot specify both only_catch_importerror and catch_exceptions',
        ):
            mod, avail = attempt_import(
                'pyomo.common.tests.dep_mod_except',
                defer_import=True,
                only_catch_importerror=True,
                catch_exceptions=(ImportError,),
            )

    def test_generate_warning(self):
        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod_except',
            defer_import=True,
            only_catch_importerror=False,
        )

        # Test generate warning
        log = StringIO()
        dep = StringIO()
        with LoggingIntercept(dep, 'pyomo.common.tests'):
            with LoggingIntercept(log, 'pyomo.common'):
                mod.generate_import_warning()
        self.assertIn(
            "The pyomo.common.tests.dep_mod_except module "
            "(an optional Pyomo dependency) failed to import",
            log.getvalue(),
        )
        self.assertIn(
            "DEPRECATED: use :py:class:`log_import_warning()`", dep.getvalue()
        )

        log = StringIO()
        dep = StringIO()
        with LoggingIntercept(dep, 'pyomo'):
            with LoggingIntercept(log, 'pyomo.core.base'):
                mod.generate_import_warning('pyomo.core.base')
        self.assertIn(
            "The pyomo.common.tests.dep_mod_except module "
            "(an optional Pyomo dependency) failed to import",
            log.getvalue(),
        )
        self.assertIn(
            "DEPRECATED: use :py:class:`log_import_warning()`", dep.getvalue()
        )

    def test_log_warning(self):
        mod, avail = attempt_import(
            'pyomo.common.tests.dep_mod_except',
            defer_import=True,
            only_catch_importerror=False,
        )
        log = StringIO()
        dep = StringIO()
        with LoggingIntercept(dep, 'pyomo'):
            with LoggingIntercept(log, 'pyomo.common'):
                mod.log_import_warning()
        self.assertIn(
            "The pyomo.common.tests.dep_mod_except module "
            "(an optional Pyomo dependency) failed to import",
            dep.getvalue(),
        )
        self.assertNotIn("DEPRECATED:", dep.getvalue())
        self.assertEqual("", log.getvalue())

        log = StringIO()
        dep = StringIO()
        with LoggingIntercept(dep, 'pyomo'):
            with LoggingIntercept(log, 'pyomo.core.base'):
                mod.log_import_warning('pyomo.core.base')
        self.assertIn(
            "The pyomo.common.tests.dep_mod_except module "
            "(an optional Pyomo dependency) failed to import",
            log.getvalue(),
        )
        self.assertEqual("", dep.getvalue())

        log = StringIO()
        with LoggingIntercept(dep, 'pyomo'):
            with LoggingIntercept(log, 'pyomo.core.base'):
                mod.log_import_warning('pyomo.core.base', "Custom")
        self.assertIn(
            "Custom (import raised ValueError: cannot import module)", log.getvalue()
        )
        self.assertEqual("", dep.getvalue())

    def test_importer(self):
        attempted_import = []

        def _importer():
            attempted_import.append(True)
            return attempt_import('pyomo.common.tests.dep_mod', defer_import=False)[0]

        mod, avail = attempt_import('foo', importer=_importer, defer_import=True)

        self.assertEqual(attempted_import, [])
        self.assertIsInstance(mod, DeferredImportModule)
        self.assertTrue(avail)
        self.assertEqual(attempted_import, [True])
        self.assertIs(mod._indicator_flag._module, dep_mod)

    def test_deferred_submodules(self):
        import pyomo

        pyo_ver = pyomo.version.version

        self.assertIsInstance(deps.pyo, DeferredImportModule)
        self.assertIsNone(deps.pyo._submodule_name)
        self.assertEqual(
            deps.pyo_available._deferred_submodules,
            ['.version', '.common', '.common.tests', '.common.tests.dep_mod'],
        )
        # This doesn't cause test_mod to be resolved
        version = deps.pyo.version
        self.assertIsInstance(deps.pyo, DeferredImportModule)
        self.assertIsNone(deps.pyo._submodule_name)
        self.assertIsInstance(deps.dm, DeferredImportModule)
        self.assertEqual(deps.dm._submodule_name, '.common.tests.dep_mod')
        self.assertIsInstance(version, DeferredImportModule)
        self.assertEqual(version._submodule_name, '.version')
        # This causes the global objects to be resolved
        self.assertEqual(version.version, pyo_ver)
        self.assertTrue(inspect.ismodule(deps.pyo))
        self.assertTrue(inspect.ismodule(deps.dm))

        with self.assertRaisesRegex(
            ValueError, "deferred_submodules is only valid if defer_import==True"
        ):
            mod, mod_available = attempt_import(
                'nonexisting.module',
                defer_import=False,
                deferred_submodules={'submod': None},
            )

        mod, mod_available = attempt_import(
            'nonexisting.module',
            defer_import=True,
            deferred_submodules={'submod.subsubmod': None},
        )
        self.assertIs(type(mod), DeferredImportModule)
        self.assertFalse(mod_available)
        _mod = mod_available._module
        self.assertIs(type(_mod), ModuleUnavailable)
        self.assertTrue(hasattr(_mod, 'submod'))
        self.assertIs(type(_mod.submod), ModuleUnavailable)
        self.assertTrue(hasattr(_mod.submod, 'subsubmod'))
        self.assertIs(type(_mod.submod.subsubmod), ModuleUnavailable)

    def test_UnavailableClass(self):
        module_obj, module_available = attempt_import(
            '__there_is_no_module_named_this__',
            'Testing import of a non-existent module',
            defer_import=False,
        )

        class A_Class(UnavailableClass(module_obj)):
            pass

        with self.assertRaisesRegex(
            DeferredImportError,
            "The class 'A_Class' cannot be created because a needed optional "
            r"dependency was not found \(import raised ModuleNotFoundError: No "
            r"module named '__there_is_no_module_named_this__'\)",
        ):
            A_Class()

        with self.assertRaisesRegex(
            DeferredImportError,
            "The class attribute 'A_Class.method' is not available because a "
            r"needed optional dependency was not found \(import raised "
            "ModuleNotFoundError: No module named "
            r"'__there_is_no_module_named_this__'\)",
        ):
            A_Class.method()

    @unittest.pytest.mark.mpi
    def test_mpi4py_available(self):
        from mpi4py import MPI

        self.assertTrue(bool(mpi4py_available))


if __name__ == '__main__':
    unittest.main()
